/*
* With HUJI mods.
**++
**  FACILITY:
**      NEWSUTILITY
**
**  ABSTRACT:
**      This module contain common support routines used by all other modules
**      of NEWS.
**
**  AUTHOR:
**      Geoff Huston
**
**  COPYRIGHT:
**      Copyright  1988,1989,1990,1991,1992
**
**  VERSION:
**	V6.0-1	11/12/90	glass@mgi.com
**	The changes here are confined to do_new_item.  The principal change is
**	the the fnam parameter (the pointer to a filename) is now allowed to be
**	NULL.  When fnam is not NULL, then do_new_item functions as before.  When
**	fnam is NULL, then do_new_item will call create_article when it determines
**	the destination filename and the xref header line.
**
**	V6.0-3	31/1/91		glass!@mgi.com
**	Retain status var on duplicate item record check in do_new_item
**	V6.1	 3-Feb-1992	rankin@eql.caltech.edu
**			substantial lint cleanup from gcc -Wall
**	V6.1	15-Feb-1992	rankin@eql.caltech.edu
**			Optimize group name lookups.
**	V6.1	13-Apr-92	Marc Shannon <synful@DRYCAS.CLUB.CC.CMU.EDU>
**	  - byte comparisons for speedup
**	V6.1    21-Jul-92	Bill Fenner <wcf@ecl.psu.edu>
**	  - improved error reporting
**	V6.1    17-Feb-1993  bailey@genetics.upenn.edu
**	  - changed 'unseenitems' to 'unreaditems' throughout
**	V6.1b7  15-May-1993  Charles Bailey  bailey@genetics.upenn.edu
**	  - optimized item number lookup, handling of message-IDs
**	  - changed map_items() to display true group count in grp dir list
**	V6.1b7  15-Jun-1993  bailey@genetics.upenn.edu
**	 - set record size in grprab to NEWS_GRPFIL_RSZ, not sizeof newsgrp,
**	   since newsgrp struct contains volatile data not needed in file,
**	   and zero fields in new grp struct beyond end of data in file
**	V6.1b7  20-Jun-1993  bailey@genetics.upenn.edu
**	 - fix bug in find_itm_by_num
**
**--
**/

#if vax11c
#module NEWSUTILITY "V6.0"
#endif

#define _NEWSUTILITY_C

#include "newsinclude.h"
#include "newsextern.h"
#ifdef vax11c
#include dcdef
#include stsdef
#include errno
#include psldef
#else
#include <dcdef.h>
#include <stsdef.h>
#include <errno.h>
#include <psldef.h>
#endif

static int screen_displays_setup = 0;

extern char dir_sincestr[];

extern
int str_found;

extern
int mailfile_open;

char nntp_currgroup[132];

extern
int profile_flags;

#ifndef _stdlib_h	/* old gcc header has bad prototype */
extern int sleep __ARGS((unsigned));
#endif
extern int lib$getjpi(), lib$put_output(), lib$create_dir();
extern int sys$setprv(), sys$getmsg(), sys$faol();

/*
 *  nosysprv
 *
 *  Turn off image installed sysprv - BUT if user already had sysprv,
 *  then leave it on (as there is no change in user functionality)
 */

void nosysprv()
{
  unsigned int authprivs[2], msysprv[2] ;
  int item = JPI$_PROCPRIV;

  msysprv[0] = PRV$M_SYSPRV;
  msysprv[1] = 0;
  _c$cks(lib$getjpi(&item,0,0,&authprivs[0],0,0));
  if (!(authprivs[0] & PRV$M_SYSPRV)) _c$cks(sys$setprv(0,msysprv,0,0));
  sysprv_off = 1;
}

/*
 *  sysprv
 *
 *  Turn sysprv on as a temp priv - assumes image was installed with SYSPRV
 */

void sysprv()
{
  unsigned int msysprv[2];

  msysprv[0] = PRV$M_SYSPRV;
  msysprv[1] = 0;
  _c$cks(sys$setprv(1,msysprv,0,0));
  sysprv_off = 0;
}

/*
 *  clear_err_line
 *
 *  error line management
 */

void clear_err_line()
{
  if (smg_active) smg$erase_line(&trailer_vd,c$ac(3),c$ac(1));
}

/*
 *  cli_catch
 *
 *  Condition handler for CLI signals
 */

int cli_catch(v_sigargs,v_mechargs)
    void *v_sigargs, *v_mechargs;
{
  char buf1[80], buf2[80];
  $DESCRIPTOR(desc1,buf1);
  $DESCRIPTOR(desc2,buf2);
  unsigned char otherargs[4];
  unsigned short strlen;
  unsigned long *sigargs = v_sigargs;

/* HUJI AXP start of modification */
#ifdef __ALPHA
#define CLI$_FACILITY 3
#else   /* __ALPHA */
#if vax11c
  globalvalue CLI$_FACILITY;
#else
#define CLI$_FACILITY 3
#endif
#endif  /* __ALPHA */
/* End of HUJI AXP mods */

#if GNUC_HANDLER_HACK
  if (sigargs[1] == C$_LONGJMP)
    return(vaxc$unwind_longjmp(v_sigargs,v_mechargs));
#endif
  if ($VMS_STATUS_FAC_NO(sigargs[1]) == CLI$_FACILITY) {
    sys$getmsg(sigargs[1],&strlen,&desc1,15,otherargs);
    desc1.dsc$w_length = strlen ;
    sys$faol(&desc1,&strlen,&desc2,&sigargs[3]);
    desc2.dsc$w_length = strlen ;
#if STRIP_ERR_LINES
    {
      int index;

      for ( index =0 ; index <=  desc2.dsc$w_length; index++ ) {
        if ( desc2.dsc$a_pointer[index] == ',' ) {
          index += 2;  /* skip comma and space */
          desc2.dsc$w_length -= index;
          desc2.dsc$a_pointer = &desc2.dsc$a_pointer[index];
          break;
          }; /* if */
        }; /* for */
    }; /* block */
#endif
    if (smg_active)
      smg$put_chars(&trailer_vd,&desc2,c$ac(3),c$ac(1),c$ac(1),c$ac(SMG$M_REVERSE),0,0);
    else lib$put_output(&desc2);
    return(SS$_CONTINUE);
    }
  else return(SS$_RESIGNAL);
}

/*
 *  create_keydefs
 *
 *  Create key definitions
 */

void create_keydefs()
{

#define ds(s) c$dsc(s)
#define term c$ac(SMG$M_KEY_TERMINATE)
#define tmne c$ac(SMG$M_KEY_TERMINATE+SMG$M_KEY_NOECHO)
#define gold ds("GOLD")
#define akd(b,c,d,e,f) smg$add_key_def(&keytab,b,c,d,e,f)

  FILE *fpr;

  smg$create_virtual_keyboard(&kid,0,0,0);
  smg$create_key_table(&keytab);
  akd(ds("PF1"),0,0,0,gold);
  akd(ds("PF2"),0,term,ds("HELP"),0);
  akd(ds("PF3"),0,term,ds("SKIP"),0);
  akd(ds("PF4"),gold,term,ds("SKIP/FOLLOWUP"),0);
  akd(ds("PF4"),0,term,ds("SKIP/NEWSGROUP"),0);
  akd(ds("KP0"),0,term,ds("READ/NEW"),0);
  akd(ds("KP0"),gold,term,ds("READ/HEADER/NEW"),0);
  akd(ds("KP1"),0,c$ac(0),ds("READ "),0);
  akd(ds("KP1"),gold,c$ac(0),ds("READ/HEADER "),0);
  akd(ds("KP2"),0,tmne,ds("DOWN"),0);
  akd(ds("KP2"),gold,tmne,ds("BOTTOM"),0);
  akd(ds("KP3"),0,term,ds("READ/NEXT"),0);
  akd(ds("KP3"),gold,term,ds("READ/NEXT/HEADER"),0);
  akd(ds("KP4"),0,term,ds("DIR"),0);
  akd(ds("KP5"),0,tmne,ds("UP"),0);
  akd(ds("KP5"),gold,tmne,ds("TOP"),0);
  akd(ds("KP6"),0,term,ds("DIR/REG"),0);
  akd(ds("KP7"),0,term,ds("DIR"),0);
  akd(ds("KP7"),gold,term,ds("PRINT"),0);
  akd(ds("KP8"),0,term,ds("DIR/NEW"),0);
  akd(ds("KP8"),gold,c$ac(0),ds("EXTRACT "),0);
  akd(ds("KP9"),0,term,ds("DIR/REG"),0);
  akd(ds("KP9"),gold,c$ac(0),ds("EXTRACT/ALL "),0);
  akd(ds("MINUS"),0,term,ds("DIR/ALL"),0);
  akd(ds("MINUS"),gold,c$ac(0),ds("EXTRACT/APPEND "),0);
  akd(ds("COMMA"),0,term,ds("DIR/ALL"),0);
  akd(ds("COMMA"),gold,term,ds("SELECT/NEW"),0);
  akd(ds("PERIOD"),0,term,ds("READ/NEW/FOLLOWUP"),0);
  akd(ds("PERIOD"),gold,term,ds("READ/NEW/FOLLOWUP/HEADER"),0);
  akd(ds("HELP"),0,term,ds("HELP"),0);
  akd(ds("FIND"),0,c$ac(0),ds("SEARCH "),0);
  akd(ds("INSERT_HERE"),0,term,ds("REGISTER"),0);
  akd(ds("REMOVE"),0,term,ds("DEREGISTER"),0);
  akd(ds("SELECT"),0,term,ds("SELECT"),0);
  akd(ds("SELECT"),gold,term,ds("CLOSE"),0);
  akd(ds("PREV_SCREEN"),0,tmne,ds("SCUP"),0);
  akd(ds("PREV_SCREEN"),gold,tmne,ds("TOP"),0);
  akd(ds("NEXT_SCREEN"),0,tmne,ds("SCDOWN"),0);
  akd(ds("NEXT_SCREEN"),gold,tmne,ds("BOTTOM"),0);
  akd(ds("CTRLW"),0,tmne,ds("REFRESH"),0);
  akd(ds("F10"),0,term,ds("EXIT"),0);
#if MENU_SUPPORT
  akd(ds("F11"),0,term,ds("Menu"),0);
#endif
  if ((fpr = fopen(News_ini,"r")) != 0) {
    fclose(fpr);
    smg$load_key_defs(&keytab,c$dsc(News_ini),0,c$ac(1));
    }
  c$cks(1);

#undef ds
#undef term
#undef tmne
#undef gold
#undef akd
}

/*
 *  cur_down_grp
 *
 *  Move the group cursor down by 1 newsgroup - (set the level to 1!)
 */

int cur_down_grp(n,refresh)
  int n, refresh;
{
  int ng,
      png = curr_g,
      ret_val = 0;

  set_level(1);
  if (n <= 0) return(0);

  if (!curr_g) return(0);
  if (ga[curr_g]->grp_display_indx) {
    for (ng = curr_g+1; ng <= ga_size; ++ng) {
      if (ga[ng]->grp_display_indx) {
        png = ng;
        ++ret_val;
        if (!--n) break;
        }
      }
    curr_g = png;
    if (smg_active) {
      if (g_arrow)
        smg$put_chars(&grp_vd,c$dsc("  "),&g_arrow,c$ac(1),0,0,0,0);
      g_arrow = ga[curr_g]->grp_display_indx;
      smg$put_chars(&grp_vd,c$dsc("->"),&g_arrow,c$ac(1),0,c$ac(SMG$M_REVERSE+SMG$M_BOLD),0,0);
      position_display(&grp_vd,&grp_paste,g_arrow,refresh,grp_display_size);
      }
    return(ret_val);
    }
  else return(0);
}

/*
 *  cur_down_itm
 *
 *  Move the item cursor down by n items
 */

int cur_down_itm(g,n,refresh)		int g,n,refresh;
{
  int ng,
      png,
      ret_val = 0;

  if (n <= 0) return(0);

  if ((g) && (ga[g]->grp_ia)) {
    if (!(png = ga[g]->grp_c_itm)) return(0);
    for (ng = ga[g]->grp_c_itm+1; ng <= ga[g]->grp_count; ++ng) {
      png = ng;
      ++ret_val;
      if (!--n) break;
      }
    if ((smg_active) && (ga[g]->grp_iavdsize)) {
      if (ga[g]->grp_c_itm) smg$put_chars(&ga[g]->grp_iavd,c$dsc("  "),c$ac(ga[g]->grp_c_itm),c$ac(1),0,0,0,0);
      smg$put_chars(&ga[g]->grp_iavd,c$dsc("->"),&png,c$ac(1),0,c$ac(SMG$M_REVERSE+SMG$M_BOLD),0,0);
      position_display(&ga[g]->grp_iavd,&ga[g]->grp_iapaste,png,refresh,ga[g]->grp_count);
      }
    ga[g]->grp_c_itm = png;
    if ((curr_g == g) && (news_context > 1)) {
      curr_i = png;
      set_level(2);
      }
    return(ret_val);
    }
  else return(0);
}

/*
 *  cur_set_grp
 *
 *  set the group pointer to g
 */

void
cur_set_grp(g)
    int g;
{
  curr_g = g;
  if (smg_active) {
    if ((curr_g) && (g_arrow == ga[curr_g]->grp_display_indx)) {
      position_display(&grp_vd,&grp_paste,g_arrow,0,grp_display_size);
      return;
      }
    if (g_arrow) smg$put_chars(&grp_vd,c$dsc("  "),&g_arrow,c$ac(1),0,0,0,0);
    if (curr_g) {
      g_arrow = ga[curr_g]->grp_display_indx;
      if (g_arrow) {
        smg$put_chars(&grp_vd,c$dsc("->"),&g_arrow,c$ac(1),0,c$ac(SMG$M_REVERSE+SMG$M_BOLD),0,0);
        position_display(&grp_vd,&grp_paste,g_arrow,0,grp_display_size);
        }
      }
    else g_arrow = 0;
    }
}

/*
 *  cur_set_itm
 *
 *  set the item pointer to item number i
 */

void cur_set_itm(g,i)
  int g,i;
{
  if (i < ga[g]->grp_c_itm) cur_up_itm(g,ga[g]->grp_c_itm - i,0);
  else if (i > ga[g]->grp_c_itm) cur_down_itm(g,i - ga[g]->grp_c_itm,0);
  else {
    if ((smg_active) && (ga[g]->grp_iavdsize)) {
      smg$put_chars(&ga[g]->grp_iavd,c$dsc("->"),c$ac(ga[g]->grp_c_itm),c$ac(1),0,c$ac(SMG$M_REVERSE+SMG$M_BOLD),0,0);
      position_display(&ga[g]->grp_iavd,&ga[g]->grp_iapaste,ga[g]->grp_c_itm,0,ga[g]->grp_count);
      }
    if ((curr_g == g) && (news_context > 1)) {
      curr_i = ga[g]->grp_c_itm;
      set_level(2);
      }
    }
}

/*
 *  cur_up_grp
 *
 *  Move the group cursor down by 1 newsgroup - (set the level to 1!)
 */

int
cur_up_grp(n,refresh)
  int n,refresh;
{
  int ng,
      png = curr_g,
      ret_val = 0;

  set_level(1);
  if (n <= 0) return(0);

  if (!curr_g) return(0);
  if (ga[curr_g]->grp_display_indx) {
    for (ng = curr_g-1; ng > 0 ; --ng) {
      if (ga[ng]->grp_display_indx) {
        png = ng;
        ++ret_val;
        if (!--n) break;
        }
      }
    curr_g = png;
    if (smg_active) {
      if (g_arrow) smg$put_chars(&grp_vd,c$dsc("  "),&g_arrow,c$ac(1),0,0,0,0);
      g_arrow = ga[curr_g]->grp_display_indx;
      smg$put_chars(&grp_vd,c$dsc("->"),&g_arrow,c$ac(1),0,c$ac(SMG$M_REVERSE+SMG$M_BOLD),0,0);
      position_display(&grp_vd,&grp_paste,g_arrow,refresh,grp_display_size);
      }
    return(ret_val);
    }
  else return(0);
}

/*
 *  cur_up_itm
 *
 *  Move the item cursor up by n items
 */

int
cur_up_itm(g,n,refresh)
  int g,n,refresh;
{
  int ng,
      png,
      ret_val = 0;

  if (n <= 0) return(0);

  if ((g) && (ga[g]->grp_ia) && (ga[g]->grp_count)) {
    png = ga[g]->grp_c_itm;
    for (ng = ga[g]->grp_c_itm-1; ng > 0; --ng) {
      png = ng;
      ++ret_val;
      if (!--n) break;
      }
    if ((smg_active) && (ga[g]->grp_iavdsize)) {
      if (ga[g]->grp_c_itm) smg$put_chars(&ga[g]->grp_iavd,c$dsc("  "),c$ac(ga[g]->grp_c_itm),c$ac(1),0,0,0,0);
      smg$put_chars(&ga[g]->grp_iavd,c$dsc("->"),&png,c$ac(1),0,c$ac(SMG$M_REVERSE+SMG$M_BOLD),0,0);
      position_display(&ga[g]->grp_iavd,&ga[g]->grp_iapaste,png,refresh,ga[g]->grp_count);
      }
    ga[g]->grp_c_itm = png;
    if ((curr_g == g) && (news_context > 1)) {
      curr_i = png;
      set_level(2);
      }
    return(ret_val);
    }
  else return(0);
}

/*
 *  do_newg
 *
 *  Create newsgroup if it does not already exist.
 */

unsigned int do_newg(g,prompt,use_sysfilter)
  char *g;
  int prompt, use_sysfilter;
{
  GRP savegrp;
  int new_mem = 0;
  char s[SUBJLEN];

  if ((!g) || (!*g)) return(0);

  util_cvrt(s,g);
  grprab.rab$l_kbf = s;
  grprab.rab$b_ksz = SUBJLEN;
  grprab.rab$b_krf = 0;
  grprab.rab$l_rop = RAB$M_RRL | RAB$M_NLK;
  grprab.rab$b_rac = RAB$C_KEY;

  if (!(sys$get(&grprab) & 1)) {
    int acpt;

    if (no_priv()) return(0);
    if (!auto_cre_grp) return(0);

    acpt = sys_accept_group(s);

    if (prompt || (!acpt && use_sysfilter)) {
      sprintf(err_oline,"Newsgroup: %s",s);
      if (!acpt) strcat(err_oline," (NEWS.SYS Filter rejects newsgroup name)");
      err_line(err_oline);
      status = get_input(&usr_inp_dsc,c$dsc("Create Newsgroup? [n]:"),&usr_inp_l);
      if ((!(status & 1)) || (!usr_inp_l)) return(0);
      if (toupper(*usr_inp) != 'Y') return(0);
      }
/*  else if (!acpt && use_sysfilter) return(0);  RRS */

    sprintf(itm_fname,Grp_template,util_dir(s));
    sysprv();
    if (!(lib$create_dir(c$dsc(itm_fname),0,0,0,0,0) & 1)) return(0);
    nosysprv();
    new_mem = 1;
    savegrp = newsgrp;
    grprab.rab$l_kbf = (char *) c$rfi(0);
    grprab.rab$b_ksz = 4;
    grprab.rab$b_krf = 1;
    grprab.rab$l_rop = RAB$M_WAT;
    grprab.rab$b_rac = RAB$C_KEY;
    _c$cks(sys$get(&grprab));

    ++newsgrp.grp_count;
    time((time_t *) &newsgrp.grp_entdate);
    savegrp.grp_num = ++newsgrp.grp_topnum;
    savegrp.grp_srvcache = newsgrp.grp_srvcache;
    savegrp.grp_srvproto = newsgrp.grp_srvproto;
    savegrp.grp_flags = newsgrp.grp_flags;
/*
    savegrp.grp_flags &= ~NEWS_M_NNTPCACHE;
    savegrp.grp_flags |= newsgrp.grp_flags & NEWS_M_NNTPCACHE;
*/
    strcpy(savegrp.grp_srvnode,newsgrp.grp_srvnode);
    _c$cks(sys$update(&grprab));

    newsgrp = savegrp;
    memset(((char *) &newsgrp) + NEWS_GRPFIL_RSZ, 0,
           sizeof newsgrp - NEWS_GRPFIL_RSZ);
    strncpy(newsgrp.grp_name,s,SUBJLEN);
    newsgrp.grp_topnum = 0;
    newsgrp.grp_firstnum = 1;
    newsgrp.grp_count = 0;
    time((time_t *) &newsgrp.grp_credate);
    time((time_t *) &newsgrp.grp_entdate);
    newsgrp.grp_life = newsgrp.grp_itmlife = 0;
    newsgrp.grp_reg = newsgrp.grp_unread = 0;
    newsgrp.grp_ia = 0;
    newsgrp.grp_iasize = 0;
    newsgrp.grp_iavd = 0;
    newsgrp.grp_display_indx = newsgrp.grp_iavdsize = newsgrp.grp_iapaste = 0;
    newsgrp.grp_flags = newsgrp.grp_c_itm = 0;
    newsgrp.grp_reg_text = (char *)0;
    newsgrp.grp_topic[0] = newsgrp.grp_notice[0] = '\0';
    grprab.rab$w_rsz = NEWS_GRPFIL_RSZ;
    _c$cks(sys$put(&grprab));
    }
  else if (!ga_search_name(s)) new_mem = 1;
  if (new_mem) do_new_mem_grp();
  return(newsgrp.grp_num);
}

/*
 *  do_new_group
 *
 *  Create newsgroup if it does not already exist.
 */

void do_new_group(g,prompt,retg)
  char *g;
  int prompt;
  unsigned int *retg;
{
  char locg[2048],
       *cp1 = locg,
       *cp2;

  strcpy(locg,g);
  *retg = 0;
  while ((cp2 = strchr(cp1,',')) != 0) {
    *cp2++ = '\0';
    if ((*retg = do_newg(cp1,prompt,1)) != 0) *++retg = 0;
    cp1 = cp2;
    }
  if ((*retg = do_newg(cp1,prompt,1)) != 0) *++retg = 0;
}

void create_newsgroups(g,retg)
  char *g;
  unsigned int *retg;
{
  char *locg, *cp1, *cp2;
  int prompt = 0, sacg = auto_cre_grp;

  auto_cre_grp = 1;
  strcpy((locg = malloc(strlen(g) + 1)),g);
  cp1 = locg;
  *retg = 0;
  while ((cp2 = strchr(cp1,',')) != 0) {
    *cp2++ = '\0';
    if ((*retg = do_newg(cp1,prompt,0)) != 0) *++retg = 0;
    cp1 = cp2;
    }
  if ((*retg = do_newg(cp1,prompt,0)) != 0) *++retg = 0;
  free(locg);
  auto_cre_grp = sacg;
}

void load_newsgroups(g,retg)
  char *g;
  unsigned int *retg;
{
  char *locg, *cp1, *cp2;
  int prompt = 0;

  strcpy((locg = malloc(strlen(g) + 1)),g);
  cp1 = locg;
  *retg = 0;
  while ((cp2 = strchr(cp1,',')) != 0) {
    *cp2++ = '\0';
    if ((*retg = do_newg(cp1,prompt,1)) != 0) *++retg = 0;
    cp1 = cp2;
    }
  if ((*retg = do_newg(cp1,prompt,1)) != 0) *++retg = 0;
  free(locg);
}

/*
 *  do_openi
 *
 *  Given an item record, open the associated text of the message.
 *  Use SYSPRV if WRITE access required, and change the mask to (g:re,w:re).
 */

FILE *do_openi(mode,server_name,protected)
  char *mode,*server_name;
  int protected;
{
  FILE *tmp_fpr;
  int nmask = 022, omask = 0;

  if (protected) nmask = 077;
  if (*mode == 'w') {
    omask = umask(nmask);
    sysprv();
    }
  else if (protected) sysprv();

  if (server_name) *server_name = 0;

/* HUJI - AXP - We can't open for read and set RAT or RFM */
   if(*mode == 'r')
        tmp_fpr = fopen(itm_fname, "r", "mbc=16");
   else tmp_fpr = fopen(itm_fname,mode,"mbc=16", "rat=cr","rfm=var");

  if (tmp_fpr && (server_name))
    *server_name = 1;
/* End of HUJI mods */

  if (*mode == 'w') {
    umask(omask);
    nosysprv();
    }
  else if (protected) nosysprv();
  return(tmp_fpr);
}

/*
 *  do_oitem
 *
 *  Given an item record, open the associated text of the message.
 *  Use SYSPRV if WRITE access required, and change the mask to (g:re,w:re).
 */

FILE *do_oitem(itm,mode,grp_name,server_name,protected)
  ITM *itm;
  char *mode, *grp_name, *server_name;
  int protected;
{
  sprintf(itm_fname,Itm_template,util_dir(grp_name),itm->itm_num);
  return(do_openi(mode,server_name,protected));
}

/*
 *  do_new_item
 *
 *  Create new item in newsgroup
 */

static
char nst[] = "";

unsigned int do_new_item(g,id,subj,fromstr,fnam,new_flag,skip_history,linecount)
  unsigned int *g;
  char *id, *subj, *fnam, *fromstr;
  int new_flag, skip_history, linecount;
{
  ITM savitm;
  struct stat sbuffer;
  unsigned int cre_grp[100], cre_itm[100];
  time_t cur_time;
  int ga_indx, i, g_count = 0;
  struct rms_file *ofile = 0, *ifile = 0;
  unsigned int *gptr;
  GRP_PTR gap = 0;
  char xrefline[250], ibuf[1024], mod[100], id_key[IDLEN + 4];
  extern int rms$errno;

  status = 1;
  *mod = '\0';
  if (fnam && !*fnam) {
    strcpy(no_new_item,"No input filename specified");
    return(0X10000000);
    }

  if ((!id) || (!*id)) {
    strcpy(no_new_item,"No message identifier string given");
    return(0X20000000);
    }

  if (!*g) {
    strcpy(no_new_item,"No newsgroup index value specified");
    return(0X3000000);
    }
  if (fnam) {
    sysprv();
    if (stat(fnam,&sbuffer)) {
      strcpy(no_new_item,"Cannot access input file");
      nosysprv();
      return(0X40000000);
      }

    if (!sbuffer.st_size) {
      nosysprv();
      strcpy(no_new_item,"Input file is empty");
      return(0X50000000);
      }

    if (!(ifile = rms_open(fnam,0))) {
      strcpy(no_new_item,"Cannot open input file (read access)");
      nosysprv();
      return(0X60000000);
      }
    nosysprv();
    }
  util_idcpy(id_key,id);

  if ((!skip_history) && (hist_check(id_key))) {
    strcpy(no_new_item,"Cannot add to Newsitem index file");
    if (ifile) rms_close(ifile);
    return(0xb0010001);
    }

  if  (!subj) subj = nst;
  sprintf(xrefline,"Xref: %s",news_node);
  time(&cur_time);
  newsitm.itm_recvdate = cur_time;
  newsitm.itm_flags = (new_flag ? NEWS_M_NEW : 0) | NEWS_M_UNREAD | NEWS_M_LINESVALID;
  newsitm.itm_lines = linecount;
  newsitm.itm_life = mail_add_expiry;
  newsitm.itm_cachedate = cur_time;
  util_subjcpy(newsitm.itm_title,subj);
  util_idcpy(newsitm.itm_id,id);
  util_fromcpy(newsitm.itm_from,fromstr);

  itmrab.rab$b_krf = 1;
  itmrab.rab$l_kbf = id_key;
  itmrab.rab$b_ksz = IDLEN + 4;
  itmrab.rab$l_rop = RAB$M_WAT;
  itmrab.rab$b_rac = RAB$C_KEY;
  itmrab.rab$w_rsz = sizeof newsitm;

  grprab.rab$b_ksz = 4;
  grprab.rab$b_krf = 1;
  grprab.rab$l_rop = RAB$M_WAT;
  grprab.rab$b_rac = RAB$C_KEY;

  for (; *g ; ++g) {
    grprab.rab$l_kbf = (char *) g;

    if (!((status = sys$get(&grprab)) & 1)) {
      status &= 0Xfffffff;
      status += 0X70000000;
      strcpy(no_new_item,"Newsgroup not located");
      continue;
      }

    if (newsgrp.grp_flags & NEWS_M_MAILGROUP) {
      sys$free(&grprab);
      status = 2;
      forward_posting = 0;
      strcpy(no_new_item,"Attempt to add item to Mail newsgroup\n");
      continue;
      }
    else if ((newsgrp.grp_flags & NEWS_M_MAILLIST) && !net_news) {
      sys$free(&grprab);
      status = 2;
      forward_posting = 0;
      strcpy(no_new_item,"Attempt to add item to MailList newsgroup\n");
      continue;
      }
    else if (newsgrp.grp_flags & NEWS_M_MAILMODERATE) {
      if (!net_news) {
        sprintf(mod,"%s@%s",usr_username,Node_address);
        forward_posting = 0;
        if (!strcmp(mod,moderator_address(newsgrp.grp_name))) forward_posting = 1;
        else if (!itm_approved) {
          if (ifile) rms_close(ifile);
          if (!fnam) fnam = create_article(NULL,NULL);
          if (fnam) {
            call_mail(fnam,subj,add_transform(moderator_address(newsgrp.grp_name)),
                      "Post to newsgroup Moderator");
            sysprv();
            ifile = rms_open(fnam,0);
            nosysprv();
            sys$free(&grprab);
            status = 2;
            strcpy(no_new_item,"Non-Approved posting to Moderated Newsgroup");
            }
          else {
            sys$free(&grprab);
            strcpy(no_new_item,"Cannot access input file");
            status = 0X40000000;
            }
          continue;
          }
        }
      else {
        if (!(newsgrp.grp_flags & NEWS_M_NOAPPROVAL) && !itm_approved) {

#if UNAPPROVED_NEWS == MAIL_TO_MODERATOR
          if (ifile) rms_close(ifile);
          if (!fnam) fnam = create_article(NULL,NULL);
          if (fnam) {
            call_mail(fnam,subj,add_transform(moderator_address(newsgrp.grp_name)),
                      "Post to newsgroup Moderator");
            sysprv();
            ifile = rms_open(fnam,0);
            nosysprv();
            }

#elif UNAPPROVED_NEWS == DISCARD
          sys$free(&grprab);
          status = 2;
          strcpy(no_new_item,"Non-Approved posting to Moderated Newsgroup");
          continue;
#endif

          }
        }
      }

    if ((ga_indx = ga_locate(*g)) > 0) gap = ga[ga_indx];
    else {
      do_new_mem_grp();
      if (!(ga_indx = ga_locate(*g))) {
        sys$free(&grprab);
        status = 2;
        strcpy(no_new_item,"Newsgroup not located (in local store)");
        continue;
        }
      gap = ga[ga_indx];
      }

    newsitm.itm_grp = newsgrp.grp_num;
    gptr = (unsigned int *) &id_key[IDLEN];
    *gptr = newsitm.itm_grp;

    savitm = newsitm;

    if ((status = sys$get(&itmrab)) & 1) {	/* #3 glass@vixvax.mgi.com */
      if (!newsitm.itm_cachedate) {
        time((time_t *) &newsitm.itm_cachedate);
        sys$update(&newsitm);
        savitm.itm_num = newsitm.itm_num;
        newsitm = savitm;
        }
      else {
        status &= 0Xfffffff;
        status += 0XB0000000;
        strcpy(no_new_item,"Cannot add to Newsitem index file");
        newsitm = savitm;
        continue;
        }
      }
    else {
      newsitm = savitm;
      ++newsgrp.grp_count;
      newsgrp.grp_entdate = cur_time;
      newsitm.itm_num = ++newsgrp.grp_topnum;
      newsitm.itm_cid = newsitm.itm_num;
      itmrab.rab$l_rbf = (char *) &newsitm;
      itmrab.rab$w_rsz = sizeof newsitm;

      if (newsgrp.grp_flags & NEWS_M_IGNEXP) newsitm.itm_life = 0;

#if ITMLIFE_DEFAULTS
      if (newsitm.itm_life) {
	if (newsgrp.grp_itmlife)
	  newsitm.itm_life = min(newsitm.itm_life,newsgrp.grp_itmlife);
	else if (ga[0]->grp_itmlife)
	  newsitm.itm_life = min(newsitm.itm_life,ga[0]->grp_itmlife);
	else
	  newsitm.itm_life = min(newsitm.itm_life,EXP_TIME);
	}
#endif

      if (!((status = sys$put(&itmrab)) & 1)) {
        int ok = 0;

        itmrab.rab$b_krf = 0;
        itmrab.rab$l_kbf = (char *) &(newsitm.itm_num);
        itmrab.rab$b_ksz = 8;
        if (sys$find(&itmrab) & 1) {
          do {
            newsitm.itm_num = ++newsgrp.grp_topnum;
            } while (sys$find(&itmrab) & 1);
          itmrab.rab$l_rbf = (char *) &newsitm;
          itmrab.rab$w_rsz = sizeof newsitm;
          if ((status = sys$put(&itmrab)) & 1) ok = 1;
          }
        itmrab.rab$b_krf = 1;
        itmrab.rab$l_kbf = id_key;
        itmrab.rab$b_ksz = IDLEN + 4;
        itmrab.rab$l_rop = RAB$M_WAT;
        itmrab.rab$b_rac = RAB$C_KEY;
        itmrab.rab$w_rsz = sizeof newsitm;
        if (!ok) {
          status &= 0Xfffffff;
          status += 0XB0000000;
          strcpy(no_new_item,"Cannot add to Newsitem index file");
          continue;
          }
        }

#if ITMLIFE_DEFAULTS
      newsitm.itm_life = mail_add_expiry;
#endif

      if (newsgrp.grp_flags & NEWS_M_IGNEXP) newsitm.itm_life = mail_add_expiry;

      if (!((status = sys$update(&grprab)) & 1)) {
        itmrab.rab$l_kbf = (char *) &(newsitm.itm_num);
        itmrab.rab$b_ksz = 8;
        itmrab.rab$b_krf = 0;
        itmrab.rab$l_rop = RAB$M_WAT;
        itmrab.rab$b_rac = RAB$C_KEY;
        if (sys$find(&itmrab) & 1) sys$delete(&itmrab);

        status &= 0Xfffffff;
        status += 0XA0000000;
        strcpy(no_new_item,"Cannot update Newsgroup index file");
        continue;
        }
      gap->grp_topnum = newsgrp.grp_topnum;
      savitm.itm_num = newsitm.itm_num;
      do_new_mem_itm(ga_indx,gap);
      newsitm.itm_num = savitm.itm_num;
      }
    cre_itm[g_count] = newsitm.itm_num;
    cre_grp[g_count++] = ga_indx;
    sprintf(&xrefline[strlen(xrefline)]," %s:%d",gap->grp_name,newsitm.itm_num);
    }

  *itm_fname = '\0';
  if (!g_count) {
    if (ifile) rms_close(ifile);
    return((!status) ? (strcpy(no_new_item,"No Valid Newsgroups"),status = 1) : status);
    }

  for (i = 0 ; i < g_count; ++i) {
    forward_posting = 1;
    if (!*itm_fname) {
      int header = 1;

      sprintf(itm_fname,Itm_template,util_dir(ga[cre_grp[i]]->grp_name),cre_itm[i]);
      if (ifile) {
        if (!(ofile = rms_create(itm_fname,0xa000))) {
          sprintf(no_new_item,"Cannot open output file (write access - %s/0x%x)",
		strerror(errno,vaxc$errno),rms$errno);
          status = 0X90000000;
          *itm_fname = '\0';
          continue;
          }
        while (rms_get(ibuf,1023,ifile)) {
          if (header && (!*ibuf)) {
            if ((g_count > 1) && (header)) rms_put(xrefline,ofile);
            rms_put(ibuf,ofile);
            header = 0;
            }
          else if (header && (!strncmp(ibuf,"Xref:",5))) {
            if (g_count > 1) rms_put(xrefline,ofile);
            header = 0;
            }
          else rms_put(ibuf,ofile);
          }
        rms_close(ifile); ifile = 0;
        rms_close(ofile);
        }
      else {
        if (create_article(itm_fname,(g_count > 1 ? xrefline : NULL)) == NULL) {
          sprintf(no_new_item,"Cannot open output file (write access - %s/0x%x)",
		strerror(errno,vaxc$errno),rms$errno);
          status = 0X90000000;
          continue;
          }
        }
      if (ga[cre_grp[i]]->grp_flags & NEWS_M_RESTRICT_SET) {
        sysprv();
        chmod(itm_fname,0700);
        nosysprv();
        }
      status = 0;
      strcpy(no_new_item,"Newsitem added");
      no_more_news = 0;
      }
    else {
      char fline[256];

      sprintf(fline,Itm_template,util_dir(ga[cre_grp[i]]->grp_name),cre_itm[i]);
      sysprv();
      file_copy(itm_fname,fline);
      if (ga[cre_grp[i]]->grp_flags & NEWS_M_RESTRICT_SET) chmod(fline,0700);
      nosysprv();
      status = 0;
      strcpy(no_new_item,"Newsitem added");
      no_more_news = 0;
      }
    }
  if (ifile) rms_close(ifile);
  return(status);
}

/*
 *  do_new_mem_grp
 *
 *  create a new group in the mem array
 */

void do_new_mem_grp()
{
  GRP_PTR nentry;
  int i,
      j;

  ++(ga[0]->grp_count);
  nentry = (GRP_PTR) malloc(sizeof newsgrp);
  *nentry = newsgrp;
  if (reorder_groups) i = ga_size + 1;
  else for (i = 1; i <= ga_size; ++i) if (strcmp(ga[i]->grp_name,newsgrp.grp_name) > 0) break;
  if (ga_malloc <= (ga_size + 2)) {
    ga_malloc += 4;
    ga = (GRP_PTR *) realloc(ga,ga_malloc * (sizeof nentry));
    }
  if (i <= ga_size) for (j = ga_size; j >= i; --j) ga[j+1] = ga[j];
  if (curr_g >= i) ++curr_g;
  ga[i] = nentry;
  ga[i]->grp_reg = NEWGROUP_REG | (1 & (profile_flags & PROFILE_NEWREGISTER));
  ga[i]->grp_display_indx = (cur_dir_type < DIR_NEW);
  ++ga_size;

  if ((gv_size) && (ga[0]->grp_count > gv_size))
    smg$change_virtual_display(&grp_vd,c$rfi(gv_size += 4),c$ac(devcol),0,0,0);
  if ((smg_active) && (cur_dir_type < DIR_NEW)) screen_map_dir();
}

/*
 *  gendate
 *
 *  generate a reasonable date string!!
 */

static
char fmt_date[26];

char *gendate(i)
  unsigned int i;
{
  if (!i) fmt_date[1] = '\0';
  else {
    strncpy(fmt_date,ctime((time_t *)&i),24);
    fmt_date[1]=fmt_date[8];fmt_date[2]=fmt_date[9];fmt_date[3]='-';
    fmt_date[7]='-';fmt_date[8]=fmt_date[22];fmt_date[9]=fmt_date[23];
    fmt_date[10]='\0';
    }
  return(&fmt_date[1]);
}

/*
 *  gen_id
 *
 *  generate a unique message id of the form <seq num>@<internet address>
 */
char genid[132];

char *gen_id()
{
  GRP savegrp;
  time_t cur_time;
  int seq = 1;
  char *timestr, *cp;

  savegrp = newsgrp;

  grprab.rab$l_kbf = (char *) c$rfi(0);
  grprab.rab$b_ksz = 4;
  grprab.rab$b_krf = 1;
  grprab.rab$l_rop = RAB$M_WAT;
  grprab.rab$b_rac = RAB$C_KEY;
  if (sys$get(&grprab) & 1) {
    seq = ++newsgrp.grp_iavd;
    sys$update(&grprab);
    }
  time(&cur_time);
  timestr = ctime(&cur_time);

  sprintf(genid,"<%.4s%.3s%.2s.%.2s%.2s%.2s.%d@%s>",
    &timestr[20],&timestr[4],&timestr[8],
    &timestr[11],&timestr[14],&timestr[17],
    seq,
    news_node);
  cp = &genid[8];
  if (*cp == ' ') {
    do *cp = *(cp + 1); while (*cp++);
    }
/* dorner@t500mo.telematik.informatik.uni-karlsruhe.de */
  if (strlen(genid) >= IDLEN) strcpy(&genid[IDLEN-2],">");
  return(genid);
}

void splashline(itmptr,vdptr,rowptr)
  ITM_PTR itmptr; int *vdptr, *rowptr;
{
  char *fdate;
  int li, ri, i;

  smg$erase_line(vdptr,rowptr,c$ac(3));

  sprintf(err_oline,"%-6d %-*.*s ",itmptr->itm_num,SUBJLEN,SUBJLEN,itmptr->itm_title);
  smg$put_chars(vdptr,c$dscl(&err_oline_d),rowptr,c$ac(3),0,0,0,0);

  if (minfromlen > 0 && !v59_file && devcol >= 80) {
    li = (67 - min(minfromlen,devcol - 30)) + (devcol - 80);
    ri = SUBJLEN + 9;
    i = min(li,ri);
    if (*(itmptr->itm_from)) {
      sprintf(err_oline," %-*.*s",FROMLEN,FROMLEN,itmptr->itm_from);
      smg$put_chars(vdptr,c$dscl(&err_oline_d),rowptr,c$ac(i),0,0,0,0);
      }
    li = devcol - 6;
    ri = i + minfromlen + 6;
    }
  else {
    li = devcol - 6;
    ri = SUBJLEN + 15;
    }
  i = max(li,ri);
  fdate = gendate(itmptr->itm_recvdate);
  if (*fdate || (itmptr->itm_flags & NEWS_M_LINESVALID)) {
    sprintf(err_oline," %-6.6s",fdate);
    smg$put_chars(vdptr,c$dscl(&err_oline_d),rowptr,c$ac(i),0,0,0,0);
    i -= 6;
    if (itmptr->itm_flags & NEWS_M_LINESVALID) {
      sprintf(err_oline,"%6d",itmptr->itm_lines);
      smg$put_chars(vdptr,c$dscl(&err_oline_d),rowptr,c$ac(i),0,0,0,0);
      }
    }
  if (itmptr->itm_flags & NEWS_M_UNREAD)
    smg$change_rendition(vdptr,rowptr,c$ac(3),c$ac(1),c$ac(6),c$ac(SMG$M_BOLD),0);
}

/*
 *  do_new_mem_itm
 *
 *  Add new item to mem array
 */

void do_new_mem_itm(g,gap)
  int g;
  GRP_PTR gap;
{
  ITM_PTR iap;
  int i;

  i = ++gap->grp_count;
  ++gap->grp_unread;
  if ((gap->grp_display_indx) && (smg_active)) {
    sprintf(err_oline,"%5d",gap->grp_count);
    smg$put_chars(&grp_vd,c$dscl(&err_oline_d),c$ac(gap->grp_display_indx),
                  c$ac(14+SUBJLEN),0,0,0,0);
    screen_update_gread(g);
    }

  if (gap->grp_ia) {
    if (gap->grp_iasize <= gap->grp_count)
      gap->grp_ia = (ITM_PTR) realloc(gap->grp_ia,(gap->grp_iasize +=4)*(sizeof newsitm));
    iap = gap->grp_ia;
    newsitm.itm_flags |= NEWS_M_UNREAD;
    gap->grp_ia[gap->grp_count] = newsitm;
    if (gap->grp_iavdsize && smg_active) {
      if (gap->grp_count > gap->grp_iavdsize)
        smg$change_virtual_display(&gap->grp_iavd,c$rfi(gap->grp_iavdsize += 4),c$ac(devcol),0,0,0);
      splashline(&(iap[i]),&(gap->grp_iavd),(int *)&(gap->grp_count));
      if (!gap->grp_c_itm) {
        gap->grp_c_itm = 1;
        smg$put_chars(&gap->grp_iavd,c$dsc("->"),c$ac(1),c$ac(1),0,c$ac(SMG$M_REVERSE+SMG$M_BOLD),0,0);
        position_display(&gap->grp_iavd,&gap->grp_iapaste,gap->grp_c_itm,0,gap->grp_count);
        if (g == curr_g) curr_i = 1;
        }
      }
    }
  else if ((smg_active) && (gap->grp_iavdsize)) {
    map_items(g);
    iap = gap->grp_ia;

    if (gap->grp_count > gap->grp_iavdsize)
      smg$change_virtual_display(&gap->grp_iavd,c$rfi(gap->grp_iavdsize += 4),
                                 c$ac(devcol),0,0,0);
    for (i = 1; i <= gap->grp_count; ++i) {
      splashline(&(iap[i]),&(gap->grp_iavd),&i);
      if (iap[i].itm_flags & NEWS_M_UNREAD) {
        if (!gap->grp_c_itm) {
          gap->grp_c_itm = i;
          smg$put_chars(&gap->grp_iavd,c$dsc("->"),&i,c$ac(1),0,c$ac(SMG$M_REVERSE+SMG$M_BOLD),0,0);
          if (g == curr_g) curr_i = i;
          }
        }
      }
    if (!gap->grp_c_itm) {
      for (i = 1; i < gap->grp_count; ++i)
      if (iap[i].itm_flags & NEWS_M_UNREAD) break;
      gap->grp_c_itm = i;
      if (gap->grp_iavdsize)
        smg$put_chars(&gap->grp_iavd,c$dsc("->"),&i,c$ac(1),0,c$ac(SMG$M_REVERSE+SMG$M_BOLD),0,0);
      position_display(&gap->grp_iavd,&gap->grp_iapaste,gap->grp_c_itm,0,gap->grp_count);
      if (g == curr_g) curr_i = i;
      }
    }
}

/*
 *  do_open_item
 *
 *  Given an item record, open the associated text of the message.
 *  Use SYSPRV if WRITE access required, and change the mask to (g:re,w:re).
 */

extern char get_server_title[];
extern char get_server_from[];
extern char get_server_id[];
extern int  get_server_size;

FILE *do_open_item(g,i,mode,server_name)
  int g,i;
  char *mode,*server_name;
{
  FILE *tmp = 0;
  char *nntp_response;
  int nntp_respval;

  if (ga[g]->grp_ia[i].itm_flags & NEWS_M_MAILITEM) {
    if (*mode == 'r') return(read_mail(g,ga[g]->grp_ia[i].itm_cid,server_name));
    else return(0);
    }

  if (nntp_client) {
    nntp_respval = 211;
    if (strcmp(ga[g]->grp_name,nntp_currgroup)) {
      sprintf(err_oline,"GROUP %s",ga[g]->grp_name);
      nntp_respval = nntp_one_call(nntp_node,nntp_proto,err_oline,&nntp_response);
      if (nntp_respval == 211) strcpy(nntp_currgroup,ga[g]->grp_name);
      }
    if (nntp_respval == 211)
      tmp = getno_server(ga[g]->grp_ia[i].itm_num,ga[g]->grp_num,nntp_node,server_name,0,nntp_proto);
    }
  else {
    sprintf(itm_fname,Itm_template,util_dir(ga[g]->grp_name),ga[g]->grp_ia[i].itm_num);
    if ((tmp = do_openi(mode,server_name,
			(ga[g]->grp_flags & NEWS_M_RESTRICT_SET))) != 0)
      return(tmp);
    if (strcmp(mode,"r") || (!(ga[g]->grp_flags & NEWS_M_NNTPSRV))) return(tmp);
    tmp = get_server_file(ga[g]->grp_ia[i].itm_id,ga[g]->grp_srvnode,server_name,
		   ((ga[g]->grp_flags & NEWS_M_NNTPCACHE) ? itm_fname : 0),
		   ga[g]->grp_srvproto,g,i);
    }
  if (tmp) {
    if (!(ga[g]->grp_ia[i].itm_flags & NEWS_M_LINESVALID)) {
      if (!nntp_client) {
        itmrab.rab$l_kbf = &(ga[g]->grp_ia[i].itm_num);
        itmrab.rab$b_ksz = 8;
        itmrab.rab$b_krf = 0;
        itmrab.rab$l_rop = RAB$M_WAT;
        itmrab.rab$b_rac = RAB$C_KEY;
        if (sys$get(&itmrab) & 1) {
          newsitm.itm_lines = get_server_size;
          newsitm.itm_flags |= NEWS_M_LINESVALID;
          util_subjcpy(newsitm.itm_title,get_server_title);
          util_fromcpy(newsitm.itm_from,get_server_from);
          if (ga[g]->grp_flags & NEWS_M_NNTPCACHE)
	    time((time_t *)&newsitm.itm_cachedate);
          sys$update(&itmrab);
          }
        util_subjcpy(ga[g]->grp_ia[i].itm_title,get_server_title);
        util_fromcpy(ga[g]->grp_ia[i].itm_from,get_server_from);
        }
      if (*get_server_id) util_idcpy(ga[g]->grp_ia[i].itm_id,get_server_id);
      ga[g]->grp_ia[i].itm_lines = get_server_size;
      ga[g]->grp_ia[i].itm_flags |= NEWS_M_LINESVALID;
      }
    util_fromcpy(ga[g]->grp_ia[i].itm_from,get_server_from);
    if (ga[g]->grp_flags & NEWS_M_NNTPCACHE)
      time((time_t *)&(ga[g]->grp_ia[i].itm_cachedate));
    else ga[g]->grp_ia[i].itm_cachedate = 0;
    if (smg_active && ga[g]->grp_iavdsize)
      splashline(&(ga[g]->grp_ia[i]),&(ga[g]->grp_iavd),&i);
    }
  return(tmp);
}

/*
 * find_itm_by_id - locate item by id
 *                  return !0 if item found, with *ip = index into grp_ia
 *                  return  0 if no item found, with *ip unchanged
 */

unsigned int find_itm_by_id(g,id,ip)
  unsigned int g, *ip;
  char *id;
{
  register unsigned int i;

  if (!map_ids(g)) return(0);
  for (i = 1; i <= ga[g]->grp_count; i++)
    if (!news_strncasecmp(ga[g]->grp_ia[i].itm_id,id,0)) return(*ip = i);
  return(0);

}

/*
 * find_itm_by_num - maps item number to ia offset
 *                   returns !0 if found, with offset in *ip
 *                   returns  0 if not exact match, with nearest offset
 *                   in *ip (iap[*ip] > itm_num preferred)
 */

unsigned int find_itm_by_num(g,itmnum,ip)
  unsigned int g, itmnum, *ip;
{                         
  static unsigned int prev_g, prev_iidx;

  if (!ga[g]->grp_ia) map_items(g);
  if (!ga[g]->grp_ia) return(*ip = 0);

  /* we expect serial lookups often, so we handle them as a special case */
  if ( (g == prev_g) && (ga[g]->grp_ia[prev_iidx+1].itm_num == itmnum) )
     return(*ip = ++prev_iidx);

  if (ga[g]->grp_flags & NEWS_M_ORDERBYSUBJECT) {
    register int i;
    for (i = 1; i <= ga[g]->grp_count; i++)
      if (ga[g]->grp_ia[i].itm_num == itmnum)
        return(prev_g = g,prev_iidx = *ip = i);
    *ip = 1;
    return(0);
    }
  else {
    register unsigned int incr, iter, tmp, maxiter = 2;
    register int cmp;

    iter = ga[g]->grp_count;
    tmp = incr = (ga[g]->grp_count > 1) ? ga[g]->grp_count / 2 : 1;
    while (iter=(iter >> 1)) maxiter++;  /* set maxiter and clear iter */
    while (1) {
      cmp = itmnum - ga[g]->grp_ia[tmp].itm_num;
      if (!cmp) return(prev_g = g,prev_iidx = *ip = tmp);
      if ( (tmp == 1) && (cmp < 0) ) {
        *ip = 1;
        return(0);
        }
      if ( (tmp == ga[g]->grp_count) && (cmp > 0) ) {
        *ip = ga[g]->grp_count;
        return(0);
        }
      if (++iter == maxiter) {
        *ip = (cmp > 0) ? ++tmp : tmp;
        return(0);
        }
      tmp += incr * ((cmp < 0) ? -1 : 1);
      if (tmp < 1) tmp = 1;
      if (tmp > ga[g]->grp_count) tmp = ga[g]->grp_count;
      incr = (++incr)/2;
      }
    }
}

static char *s_newsgroup = 0;
static int s_inum = 0;

static int snewsgroup(header_line)
  char *header_line;
{
  char *cp;

  if (!s_inum && !strncmp(header_line,"Xref:",5)) {
    header_line += 5;
    while (isspace(*header_line)) ++header_line;
    if ((cp = chop_str_plus(header_line,':')) != 0) {
      if (s_newsgroup) free(s_newsgroup);
      s_newsgroup = malloc(strlen(header_line) + 1);
      strcpy(s_newsgroup,header_line);
      lower_case(s_newsgroup);
      sscanf(cp,"%d",&s_inum);
      }
    return 0;
    }
  if (strncmp(header_line,"Newsgroups:",11)) return 0;
  chop_str(header_line,'\n');
  cp = &header_line[11];
  while (isspace(*cp)) ++cp;
  s_newsgroup = malloc(strlen(cp) + 1);
  strcpy(s_newsgroup,cp);
  lower_case(s_newsgroup);
  return 1;
}

static int sinum(resp_line)
  char *resp_line;
{
  if (sscanf(resp_line,"221 %d",&s_inum) != 1) s_inum = 0;
  return 1;
}

static int fib(header_line)
  char *header_line;
{
  char *cp;

  chop_str(header_line,'\n');
  if (!strncmp(header_line,"From:",5)) {
    cp = &header_line[6];
    while (isspace(*cp)) ++cp;
    util_fromcpy(newsitm.itm_from,cp);
    }
  else if (!strncmp(header_line,"Subject:",8)) {
    cp = &header_line[8];
    while (isspace(*cp)) ++cp;
    util_subjcpy(newsitm.itm_title,cp);
    }
  else if (!strncmp(header_line,"Date:",5)) {
    int tmp;

    cp = &header_line[6];
    while (isspace(*cp)) ++cp;
    if ((tmp = parse_usenet_date(cp)) != 0) newsitm.itm_postdate = tmp;
    }
  else if (!strncmp(header_line,"Lines:",5)) {
    if (sscanf(header_line,"Lines: %d",&newsitm.itm_lines) == 1)
      newsitm.itm_flags |= NEWS_M_LINESVALID;
    }
  return 1;
}

int get_itm_by_ident(id,rg,ri,def_g)
  char *id;
  int *rg, *ri,def_g;
{
  char pref[512], *cp1, *cp2, *cp3, *cp4, *nntp_response;
  int count, first, last, g, i;
  unsigned int *gr;

  if (!nntp_client) {
    util_idcpy(pref,id);
    gr = (unsigned int *) &(pref[IDLEN]);
    *gr = ga[curr_g]->grp_num;
    itmrab.rab$l_kbf = pref;
    itmrab.rab$b_ksz = IDLEN + 4;
    itmrab.rab$b_krf = 1;
    itmrab.rab$l_rop = RAB$M_RRL | RAB$M_NLK;
    itmrab.rab$b_rac = RAB$C_KEY;
    g = curr_g;
    if (!(sys$get(&itmrab) & 1)) {
      *gr = 0;
      itmrab.rab$l_rop = RAB$M_KGE | RAB$M_RRL | RAB$M_NLK;
      if ((!(sys$get(&itmrab) & 1)) || (strcmp(pref,newsitm.itm_id))) return(0);
      if (!(g = ga_locate(newsitm.itm_grp))) return(0);
      }
    s_inum = newsitm.itm_num;
    *rg = g;
    if (!ga[g]->grp_ia) map_items(g);
    if (!ga[g]->grp_ia) return(0);
    if (!find_itm_by_num(g,s_inum,&i)) return(0);
    *ri = i;
    return(1);
    }
  if (s_newsgroup) free(s_newsgroup);
  s_newsgroup = 0;
  s_inum = 0;
  sprintf(err_oline,"HEAD %s",id);
  if (!nntp_get_info(nntp_node,nntp_proto,err_oline,221,sinum,snewsgroup)) {
    if (def_g) {
      s_newsgroup = malloc(strlen(ga[def_g]->grp_name) + 1);
      strcpy(s_newsgroup,ga[def_g]->grp_name);
      }
    else return(0);
    }
  cp1 = s_newsgroup;
  do {
    if (cp1 == ga[curr_g]->grp_name) {
      cp2 = s_newsgroup;
      g = curr_g;
      }
    else {
      cp2 = chop_str_plus(cp1,',');
      if (!(g = ga_exact_name(cp1))) continue;
      }
    if (find_itm_by_id(g,id,ri)) return(1);
    sprintf(err_oline,"GROUP %s",cp1);
    if (nntp_one_call(nntp_node,nntp_proto,err_oline,&nntp_response) != 211) continue;
    strcpy(nntp_currgroup,cp1);
    if (sscanf(nntp_response,"211 %d %d %d",&count,&first,&last) != 3) continue;
    if (s_inum && ((s_inum < first) || (s_inum > last))) continue;
    if (!s_inum) {
      IDMAP_PTR tmp = ga[g]->grp_idhead;

      do {
        if (!news_strncasecmp(tmp->id,id,0)) {
          s_inum = tmp->itm_num;
          tmp->itm_num = (ga[g]->grp_idtail)->itm_num;
          util_idcpy(tmp->id,(ga[g]->grp_idtail)->id);
          tmp = ga[g]->grp_idtail;
          if (!(ga[g]->grp_idtail = tmp->b_link)) ga[g]->grp_idhead = 0;
          free(tmp);
          (ga[g]->grp_idtail)->f_link = 0;
          break;
          }
        } while (tmp = tmp->f_link);
      if (!s_inum) continue;
      }
    sprintf(err_oline,"STAT %d",s_inum);
    if (nntp_one_call(nntp_node,nntp_proto,err_oline,&nntp_response) != 223) continue;
    if (!(cp3 = strchr(nntp_response,'<'))) continue;
    if ((cp4 = strrchr(cp3,'>')) != 0) *++cp4 = '\0';
    if (news_strncasecmp(cp3,id,0)) continue;
    *rg = g;
    if (!ga[g]->grp_ia) map_items(g);
    if (!ga[g]->grp_ia) continue;
    if (!find_itm_by_num(g,s_inum,&i)) {
      int sav_inum = 0, loc_i;
      GRP_PTR gap = ga[g];

      if ((curr_g == g) && (news_context > 1)) {
        if (curr_i) sav_inum = gap->grp_ia[curr_i].itm_num; else sav_inum = -1;
        set_level(1);
        }
      if (gap->grp_iavdsize) smg$delete_virtual_display(&gap->grp_iavd);
      gap->grp_iavdsize = 0;
      sprintf(err_oline,"HEAD %d",s_inum);
      newsitm.itm_num = s_inum;
      newsitm.itm_grp = g;
      util_idcpy(newsitm.itm_id,id);
      time((time_t *)&newsitm.itm_recvdate);
      newsitm.itm_lines = 0;
      newsitm.itm_title[0] = '\0';
      newsitm.itm_cachedate = 0;
      newsitm.itm_flags = 0;
      newsitm.itm_cid = 0;
      newsitm.itm_from[0] = '\0';
      if (!nntp_get_info(nntp_node,nntp_proto,err_oline,221,0,fib)) continue;
      newsitm.itm_postdate = newsitm.itm_recvdate;
      if (!gap->grp_ia) {
        gap->grp_iasize = gap->grp_count + 2;
        gap->grp_ia = malloc((gap->grp_iasize + 1) * (sizeof newsitm));
        gap->grp_count = 0;
        }
      else if (gap->grp_iasize <= gap->grp_count)
        gap->grp_ia = (ITM_PTR) realloc(gap->grp_ia,(gap->grp_iasize +=4)*(sizeof newsitm));
      ++gap->grp_count;
      for (i = gap->grp_count; i  > 1; --i) {
        if (gap->grp_ia[i-1].itm_num > s_inum) gap->grp_ia[i] = gap->grp_ia[i-1];
        else {
          gap->grp_ia[i] = newsitm;
          break;
          }
        }
      if (i == 1) gap->grp_ia[1] = newsitm;
      if (sav_inum) {
	set_level(2);
        if ( (sav_inum > 0) && find_itm_by_num(g,sav_inum,&loc_i) )
          cur_set_itm(g,loc_i);
        }
      }
    *ri = i;
    return(1);
    } while ((cp1 = cp2) != 0);
  return(0);
}

/*
 *  do_open_header
 *
 *  Given an item record, open the associated header of the message.
 */

FILE *do_open_header(g,i,mode,server_name)
  int g,i;
  char *mode,*server_name;
{
  FILE *tmp;

  if (!nntp_client) {
    sprintf(itm_fname,Itm_template,util_dir(ga[g]->grp_name),ga[g]->grp_ia[i].itm_num);
    if ((tmp = do_openi(mode,server_name,
			(ga[g]->grp_flags & NEWS_M_RESTRICT_SET))) != 0)
      return(tmp);
    if (strcmp(mode,"r") || (!(ga[g]->grp_flags & NEWS_M_NNTPSRV)))
      return(tmp);
    }
  tmp = get_server_header(ga[g]->grp_ia[i].itm_id,ga[g]->grp_srvnode,server_name,
			  ga[g]->grp_srvproto,ga[g]->grp_name,ga[g]->grp_ia[i].itm_num);
  if (tmp) {
    if (!(ga[g]->grp_ia[i].itm_flags & NEWS_M_LINESVALID)) {
      itmrab.rab$l_kbf = (char *) &(ga[g]->grp_ia[i].itm_num);
      itmrab.rab$b_ksz = 8;
      itmrab.rab$b_krf = 0;
      itmrab.rab$l_rop = RAB$M_WAT;
      itmrab.rab$b_rac = RAB$C_KEY;
      if (sys$get(&itmrab) & 1) {
        newsitm.itm_lines = get_server_size;
        newsitm.itm_flags |= NEWS_M_LINESVALID;
        util_subjcpy(newsitm.itm_title,get_server_title);
        util_fromcpy(newsitm.itm_from,get_server_from);
        sys$update(&itmrab);
        }
      ga[g]->grp_ia[i].itm_lines = get_server_size;
      ga[g]->grp_ia[i].itm_flags |= NEWS_M_LINESVALID;
      util_subjcpy(ga[g]->grp_ia[i].itm_title,get_server_title);
      util_fromcpy(ga[g]->grp_ia[i].itm_from,get_server_from);
      }
    if (smg_active && ga[g]->grp_iavdsize)
      splashline(&(ga[g]->grp_ia[i]),&(ga[g]->grp_iavd),&i);
    }
  return(tmp);
}

/*
 *  check_command
 *
 *  command parser without dispatch
 */

static char s_param[256];

static int itf()
{
  if (news_context > 1) {
    int i;

    if (!curr_g) return(0);
    if (!(i = item_find(s_param))) return(0);
    cur_set_itm(curr_g,i);
    return(0);
    }
  else if (smg_active && curr_g && ga[curr_g]->grp_display_indx) {
    int num = 1;

    sscanf(s_param,"%d",&num);
    if (num < ga[curr_g]->grp_display_indx) cur_up_grp(ga[curr_g]->grp_display_indx - num,1);
    else if (num > ga[curr_g]->grp_display_indx) cur_down_grp(num - ga[curr_g]->grp_display_indx,1);
    return(0);
    }
  else return(do_error(),0);
}

static int dcitm()
{
  if (curr_i <= 0) return(do_error(),0);
  return(do_readfind(s_param,0,0),0);
}

int chk_val = -1;

int
check_command(s)
  char *s;
{
  int i;

  while (*s && isspace(*s)) s++;
  i = strlen(s);
  while (i && isspace(s[i-1])) --i;
  s[i] = '\0';

  if ((*s == '!') || (*s == ';')) return(chk_val = 0);
  if (!*s) return(chk_val = 1);
  if (isdigit(*s)) return(chk_val = 2);
  if (!strcmp(s,".")) return(chk_val = 3);
  if (cli$dcl_parse(c$dsc(s),CLICMDTBL,get_input,get_input,c$dsc("NEWS> ")) & 1)
    return(chk_val = 4);
  return(chk_val = 0);
}
/*
 *  do_parse
 *
 *  command parser
 */

int
do_parse(s)
  char *s;
{
  int parse_val, i;

  if ((parse_val = chk_val) == -1) parse_val = check_command(s);
  chk_val = -1;
  if (!parse_val) return(0);

  while (*s && isspace(*s)) s++;
  i = strlen(s);
  while (i && isspace(s[i-1])) --i;
  s[i] = '\0';

  if (parse_val == 1) {
    /* This procedure is called from two locations: at the main parse loop and within
     * the text display loop. If in the text display loop, then parse is not called
     * with an empty string - hence the parse_val does not need to consider
     * stack unwind operations
     */
    if ((news_context == 1) && (curr_g)) return(do_select("%",curr_g,3),0);

    else if (old_context == 3) {
      if (*profile_endofitm_cmd) return(do_parse(profile_endofitm_cmd));
      else return(do_readnext(0,0,0),0);
/*
           return(do_readfollow(0,0,1),0);
 */
      }
    else {
      if ((curr_g) && (curr_i)) {
        item_update(curr_g,curr_i);
        rpush(curr_g,curr_i);
        do_display(0,0,0);
        }
      return(0);
      }
    }
  if (parse_val == 2) {
    /* The input is a digit - this is a stack unwind operation */
    strcpy(s_param,s);
    if (news_context > 2) return(unwind_display(OUTER_LOOP,dcitm));
    return(unwind_display(OUTER_LOOP,itf));
    }
  if (parse_val == 3) {
    /* The input is a '.' - this is a stack unwind operation */
    strcpy(s_param,s);
    return(unwind_display(OUTER_LOOP,dcitm));
    }
  if (parse_val == 4) return(cli$dispatch());
  return(0);
}

void
write_err_line(s)
    char *s;
{
  err_line(s);
  if (smg_active) {
    smg$end_pasteboard_update(&pid);
    smg$begin_pasteboard_update(&pid);
    }
}

/*
 *  ga_locate
 *
 *  locate newsgroup index from newsgroup num
 */

int
ga_locate(g)
  unsigned int g;
{
  int i;

  for (i = 1; i <= ga_size; ++i)
    if (g == ga[i]->grp_num) return(i);
  return(0);
}

/*
 *  ga_search_name
 *
 *  locate a newsgroup by name
 */

int ga_search_name(s)
    char *s;
{
  int j = 0, i, l = strlen(s);
  char c = *s;

  for (i = 1 ; i <= ga_size ; ++i) {
    if (*ga[i]->grp_name != c) continue;
    if (!strcmp(s,ga[i]->grp_name)) return(i);
    if (!strncmp(s,ga[i]->grp_name,l)) j = i;
    }
  return(j);
}

/*
 *  ga_exact_name
 *
 *  locate a newsgroup by name - exact match
 */


static int ga_exact_compare(v_key, v_test)
  const void *v_key, *v_test;
{
  const char *key = v_key;
  const GRP_PTR *test = v_test;

  return strcmp(key, (*test)->grp_name);
}
  
int ga_exact_name(s)
  char *s;
{
  static int prev_gi = 0;
  GRP_PTR *ga_i;
  int i, slen;

  /* chance of consecutive lookup attempt is high enough to make a special
     check worthwhile
   */
  if (++prev_gi <= ga_size && !strcmp(s,ga[prev_gi]->grp_name))
    return prev_gi;

  if ( ! reorder_groups ) {  /* perform binary search */
    ga_i = bsearch(s, ga+1, ga_size, sizeof *ga, ga_exact_compare);
    return prev_gi = (ga_i ? (ga_i - ga) : 0);
  }
/* else use old code. */
  slen = strlen(s);
  if (slen >= SUBJLEN)
      slen = SUBJLEN - 1;
  slen = (slen > 0) ? slen - 1 : 0;

  /* This little trick lets us do single byte compares on each newsgroup.
     Assuming that the group for which we are searching has a name of less than
     SUBJLEN characters (#define'd maximum), then slen should be a valid index
     into the character array pointing at the last character.  This essentially
     checks to see if the two group names have the same length. */

  for (i = 1; i <= ga_size; ++i)
    if (s[slen] == ga[i]->grp_name[slen])
        if (!strcmp(s, ga[i]->grp_name))
            return (i);
  return(0);
}

/*
 *  create_pb
 *
 *  create the pastebaord
 */

struct {
    unsigned long int SMG$L_DEVCHAR;
    unsigned long int SMG$L_DEVDEPEND;
    unsigned long int SMG$L_DEVDEPEND2;
    unsigned char SMG$B_DEVCLASS;
    unsigned char SMG$B_SMG_DEVTYPE;
    unsigned char SMG$B_PHY_DEVTYPE;
    unsigned char SMG$B_ROWS;
    unsigned short int SMG$W_WIDTH;
    unsigned char SMG$B_COLOR;
    unsigned char SMG$B_PARITY;
    unsigned short int SMG$W_SPEED;
    unsigned short int SMG$W_FILL;
    unsigned short int SMG$W_CURSOR_ROW;
    unsigned short int SMG$W_CURSOR_COL;
    unsigned long int SMG$L_CURSOR_DID;
    } pb_info;

static int smg_term = -1;

static struct bmsg {
  char *bm;
  struct bmsg *bnext;
  } *bhead = 0, *btail = 0;;

static int bcount = 0,
           bbytecount = 0;

int line_width = 80;

int
do_clear_messages()
{
  err_line("Message Buffer cleared");
  while (bhead) {
    btail = bhead ;
    bhead = bhead->bnext;
    free(btail->bm);
    free(btail);
    }
  btail = 0;
  bcount = bbytecount = 0;
  return(0);
}

void push_brdmsg(str)
  char *str;
{
  int i = strlen(str);

  if (!btail) bhead = btail = malloc(sizeof *btail);
  else {
    btail->bnext = malloc(sizeof *btail);
    btail = btail->bnext;
    }
  btail->bnext =0;
  bbytecount += i;
  strcpy((btail->bm = malloc(++i)),str);
  ++bcount;
}

/*
 *  err_line
 *
 *  error line management
 */

void
err_line(s)
    char *s;
{
#if STRIP_ERR_LINES
  if (!strncmp(s,"Error",5) || !strncmp(s,"\tError",6)) {
    char *cp = &s[6];

    while (*cp && !isalnum(*cp)) cp++;
    if (*cp) s = cp;
    }
#endif
  push_brdmsg(s);
  if (!smg_active) printf("\t%s\n",s);
  else smg$put_chars(&trailer_vd,c$dsc(s),c$ac(3),c$ac(1),c$ac(1),c$ac(SMG$M_REVERSE),0,0);
}

extern struct stat dispf_stat;
extern int gets_size,
           text_lines,
           text_cline;

static int
dsm_1()
{
  if (!bhead) {
    strcpy(err_oline,"Show Messages - message buffer is empty");
    if (!smg_active) printf("\t%s\n",err_oline);
    else smg$put_chars(&trailer_vd,c$dscl(&err_oline_d),c$ac(3),c$ac(1),c$ac(1),c$ac(SMG$M_REVERSE),0,0);
    return(0);
    }
  textptr.str_searching = 1;
  strcpy(textptr.str_target,"#### END of MESSAGE BUFFER ####");
  textptr.str_length = strlen(textptr.str_target);
  textptr.str_case_sensitive = 1;
  start_header(T_DISPLAY_LOOP,SHOW_MESSAGES);
  textptr.str_found = 0;
  sprintf(err_oline,"Message Buffer           Total size: %d messages",bcount);
  put_line(err_oline,0,T_DISPLAY_LOOP);
  end_header(T_DISPLAY_LOOP);
  textptr.text_cline = 0;
  textptr.text_lines = bcount;
  textptr.dispf_stat.st_size = bbytecount;
  btail = bhead;
  for (;;) {
    textptr.gets_size += strlen(btail->bm);
    put_line(btail->bm,0,T_DISPLAY_LOOP);
    if (!btail->bnext) break;
    btail = btail->bnext;
    }
  put_line("#### END of MESSAGE BUFFER ####",0,T_DISPLAY_LOOP);
  put_line("",1,T_DISPLAY_LOOP);
  textptr.str_searching = 0;
  return(0);
}

int
do_show_messages()
{
  return(unwind_display(I_DISPLAY_LOOP,dsm_1));
}

void display_brdcst(line)	int line;
{
  char bm[512];
  unsigned short bm_len;
  $DESCRIPTOR(bm_dsc,bm);
  int sts;

  if (((sts = smg$get_broadcast_message(&pid,&bm_dsc,&bm_len,0)) & 1)
      && (sts != SMG$_NO_MORMSG)) {
    bm[bm_len] = '\0';
    smg$erase_line(&trailer_vd,&line,c$ac(1));
    smg$put_chars(&trailer_vd,c$dsc(bm),&line,c$ac(1),c$ac(1),c$ac(SMG$M_BOLD),0,0);
    smg$set_cursor_abs(&trailer_vd,c$ac(1),c$ac(brdcst_col));
    push_brdmsg(bm);
    }
}

void b_trap()
{
  if (brdcst_line) display_brdcst(brdcst_line);
}

int create_pb()
{
  if (pid_created) return(1);
  if (smg_term >= 0) return(smg_term);
  smg$create_pasteboard(&pid,c$dsc("SYS$OUTPUT"),&devrow,&devcol,c$ac(0));
  pid_created = 1;
  _c$cks(smg$get_pasteboard_attributes(&pid,&pb_info,c$ac(sizeof pb_info)));
  if (pb_info.SMG$B_DEVCLASS != DC$_TERM) {
    printf("\n\tNEWS: Not a terminal - /NOSCREEN mode used\n");
    pid_created = 0;
    }
  else if (pb_info.SMG$B_SMG_DEVTYPE != SMG$K_VTTERMTABLE) {
    printf("\n\tNEWS: Not an SMG supported terminal - /NOSCREEN mode used\n");
    pid_created = 0;
    }
  line_width = pb_info.SMG$W_WIDTH;
  if (!pid_created) smg$delete_pasteboard(&pid,c$ac(0));
  return(smg_term = pid_created);
}

/*
 *  init_screen
 *
 *  set up the screen display
 */

int init_screen()
{
  int num_grps;

  if (smg_active) return(0);
  if (!create_pb()) {
    printf("\nError: SCREEN - Not an SMG terminal\n");
    return(0);
    }
  set_level(1);
  smg_active = 1;
  num_grps = ga[0]->grp_count;
  c$cks(smg$create_virtual_display(c$rfi(num_grps + 40),c$ac(devcol),&grp_vd,0,0,0));
  gv_size = num_grps + 40;
  c$cks(smg$create_virtual_display(c$ac(3),c$ac(devcol),&grp_header_vd,0,c$ac(SMG$M_REVERSE + SMG$M_BOLD),0));
  c$cks(smg$create_virtual_display(c$ac(3),c$ac(devcol),&itm_header_vd,0,c$ac(SMG$M_REVERSE + SMG$M_BOLD),0));
  c$cks(smg$create_virtual_display(c$ac(3),c$ac(devcol),&trailer_vd,0,0,0));
  c$cks(smg$create_virtual_display(c$rfi(DISP_BUF_SIZE + devrow),c$ac(devcol),&(itmptr.sdid),0,0,0));
  c$cks(smg$create_virtual_display(c$rfi(DISP_BUF_SIZE + devrow),c$ac(devcol),&(textptr.sdid),0,0,0));
  smg$set_display_scroll_region(&(itmptr.sdid),c$ac(1),c$ac(DISP_BUF_SIZE));
  smg$set_display_scroll_region(&(textptr.sdid),c$ac(1),c$ac(DISP_BUF_SIZE));
  c$cks(smg$create_virtual_display(c$rfi(devrow - 3),c$ac(devcol),&(itmptr.shdid),0,0,0));
  c$cks(smg$create_virtual_display(c$rfi(devrow - 3),c$ac(devcol),&(textptr.shdid),0,0,0));

  smg$put_chars(&grp_header_vd,c$dsc("NEWS       NEWSGROUP LISTING"),c$ac(1),c$ac(4),0,0,0,0);
  smg$put_chars(&grp_header_vd,c$dsc("        Newsgroup                                           Count    Unread"),c$ac(3),c$ac(4),0,0,0,0);

  smg$put_chars(&itm_header_vd,c$dsc("NEWS ITEMS"),c$ac(1),c$ac(4),0,0,0,0);
  smg$put_chars(&itm_header_vd,c$dsc("      Title"),c$ac(3),c$ac(4),0,0,0,0);

  smg$begin_pasteboard_update(&pid);
  smg$erase_pasteboard(&pid);
  do_refresh();
  c$cks(smg$paste_virtual_display(&grp_vd,&pid,c$ac(4),c$ac(1)));
  c$cks(smg$paste_virtual_display(&grp_header_vd,&pid,c$ac(1),c$ac(1)));
  c$cks(smg$paste_virtual_display(&trailer_vd,&pid,c$rfi(devrow - 2),c$ac(1)));

  screen_map_dir();

  chg_mode();
  if (broadcast_trapping_requested) broad_trap(); else no_broad_trap();
  screen_displays_setup = 1;
  return(0);
}

int do_screen()
{
  return(unwind_display(OUTER_LOOP,init_screen));
}

/*
 *  item_find
 *
 *  locate a news item as specified by item number
 */

int item_find(s)
char *s;
{
  int locnum,
      i;
  ITM_PTR iap;

  if (!curr_g) {
    err_line("No Group Selected");
    return(0);
    }

  if (!ga[curr_g]->grp_count) return(0);

  if ((s) && (isdigit(*s))) {
    set_level(2);
    if (sscanf(s,"%d",&locnum) != 1) return(0);
    if (locnum <= ga[curr_g]->grp_ia[1].itm_num) {
      cur_up_itm(curr_g,ga[curr_g]->grp_count,0);
      return(curr_i);
      }
    iap = ga[curr_g]->grp_ia;
    find_itm_by_num(curr_g,locnum,&i);
    cur_set_itm(curr_g,i);
    return(curr_i);
    }
  if ((s) && (!strcmp(s,"*"))) {
    set_level(2);
    cur_set_itm(curr_g,ga[curr_g]->grp_count);
    return(curr_i);
    }
  return(0);
}

/*
 *  map_items
 *
 *  Map items into the mem array
 */

#if NNTP_CLIENT
static GRP_PTR sgap;
static unsigned int sgi;

static int initni(cp1)
  char *cp1;
{
  char *cp3;
  int indx, i, j;

  chop_str(cp1,'\n');
  if (sscanf(cp1,"%d",&indx) != 1) return 0;

  if (sgap->grp_count && find_itm_by_num(sgi,indx,&i)) return 0;
  else i = sgap->grp_count + 1;

  cp3 = cp1;
  while (!isspace(*cp3)) ++cp3;
  while (isspace(*cp3)) ++cp3;

  if (sgap->grp_iasize == sgap->grp_count) {
    sgap->grp_iasize += 10;
    sgap->grp_ia = realloc(sgap->grp_ia,(sgap->grp_iasize + 1) * (sizeof newsitm));
    }
  j = sgap->grp_count + 1;
  while (j > i) {
    sgap->grp_ia[j] = sgap->grp_ia[j-1];
    --j;
    }
  newsitm.itm_num = indx;
  newsitm.itm_grp = sgap->grp_num;
  *(newsitm.itm_id) = '\0';
  newsitm.itm_recvdate = 0;
  newsitm.itm_lines = 0;
  util_subjcpy(newsitm.itm_title,cp3);
  newsitm.itm_cachedate = 0;
  newsitm.itm_life = 0;
  newsitm.itm_flags = NEWS_M_UNREAD;
  newsitm.itm_cid = 0;
  sgap->grp_ia[i] = newsitm;
  ++sgap->grp_count;
  return 1;
}
#endif

int map_items(g)
  unsigned int g;
{
  unsigned int key[2];
  GRP_PTR gap;
  ITM_PTR l_ia;
  char *itm_list, *bufptr, *cp1, *cp2, *cp3, subject[256];
  int f_itm, l_itm, i, j, count, first, last, indx, xhdr_call = 0, loc_disp_us,
      loc_disp_ui, loc_disp_stk;

  if ((!g) || (ga[g]->grp_ia)) return 0;
  loc_disp_us = display_unseen_stack;
  loc_disp_stk = display_stk;
  loc_disp_ui = display_unseen_items;
  gap = ga[g];
  if (gap->grp_flags & NEWS_M_NONDEFDISPLAY) {
    loc_disp_us = gap->grp_flags & NEWS_M_UNSEENSTACK;
    if (loc_disp_us) loc_disp_stk = gap->grp_displaystk;
    loc_disp_ui = gap->grp_flags & NEWS_M_UNREADITEMS;
    }
  if (loc_disp_ui) {
    loc_disp_us = 1;
    loc_disp_stk = 0;
    }

  gap->grp_iasize = gap->grp_count + 2;
  gap->grp_ia = malloc((gap->grp_iasize + 1) * (sizeof newsitm));
  gap->grp_count = 0;

  if (gap->grp_flags & NEWS_M_MAILGROUP) return(mailmap(g));

  if (nntp_client) {
    *nntp_currgroup = '\0';
    sprintf(err_oline,"GROUP %s",gap->grp_name);
    if (nntp_one_call(nntp_node,nntp_proto,err_oline,&bufptr) == 211) {
      count = first = last = 0;
      sscanf(bufptr,"211 %d %d %d",&count,&first,&last);
      if (smg_active && gap->grp_display_indx) {
        sprintf(err_oline,"%5d",count);
        smg$put_chars(&grp_vd,c$dscl(&err_oline_d),c$ac(gap->grp_display_indx),c$rfi(14+SUBJLEN),0,0,0,0);
        }
      if (count && gap->grp_reg_text && loc_disp_us) {
        char *cp;
        int funr;

        if ((cp = strchr(gap->grp_reg_text,'<')) && (sscanf(cp+1,"%d",&funr) == 1)) {
          funr++;
          funr -= loc_disp_stk;
          if (funr > first) first = funr;
          if (first > last) count = 0;
          }
        if ((cp = strchr(gap->grp_reg_text,'-')) && (sscanf(cp+1,"%d",&funr) == 1)) {
          funr++;
          funr -= loc_disp_stk;
          if (funr > first) first = funr;
          if (first > last) count = 0;
          }
        }
      if (count) {
        sgap = gap;
        sgi = g;
        sprintf(err_oline,"XHDR Subject %d-%d",first,last);
        if ((NNTP_USE_XHDR) &&
            (nntp_get_info(nntp_node,nntp_proto,err_oline,221,0,initni))) {
          if (profile_display_lines) ds_lines(g);
          if (profile_display_postmark) ds_postmark(g);
          if (minfromlen > 0) ds_from(g);
          xhdr_call = 1;
          }
        else {
          for (indx = first; indx <= last; ++indx) {
            sprintf(err_oline,"HEAD %d",indx);
            if (nntp_call(nntp_node,nntp_proto,err_oline,&bufptr) == 221) {
              newsitm.itm_grp = gap->grp_num;
              util_subjcpy(newsitm.itm_title,"--Title not retrieved--");
              newsitm.itm_recvdate = 0;
              newsitm.itm_lines = 0;
              newsitm.itm_cachedate = 0;
              newsitm.itm_life = 0;
              newsitm.itm_flags = NEWS_M_UNREAD;	/* glass@mgi.com  V6.0-3  4/2/91 */
              newsitm.itm_cid = 0;
	      cp1 = chop_str_plus(bufptr,'\n');
              sscanf(bufptr,"221 %d %s",&newsitm.itm_num,subject);
              util_idcpy(newsitm.itm_id,subject);
              while (cp1) {
		cp2 = chop_str_plus(cp1,'\n');
		if ((cp3 = chop_str_plus(cp1,':')) != 0) {
                  while (isspace(*cp3)) ++cp3;
                  lower_case(cp1);
                  if (!strcmp(cp1,"subject")) util_subjcpy(newsitm.itm_title,cp3);
                  else if (!strcmp(cp1,"from")) util_fromcpy(newsitm.itm_from,cp3);
                  else if (!strcmp(cp1,"message-id")) util_idcpy(newsitm.itm_id,cp3);
                  else if (!strcmp(cp1,"lines")) {
                    if (sscanf(cp3,"%d",&(newsitm.itm_lines)) == 1)
                      newsitm.itm_flags |= NEWS_M_LINESVALID;
                    }
                  else if (!strcmp(cp1,"date"))
                    newsitm.itm_postdate = newsitm.itm_recvdate = parse_usenet_date(cp3);
		  }
                cp1 = cp2;
                }
              if (gap->grp_iasize == gap->grp_count) {
                gap->grp_iasize += 10;
                gap->grp_ia = realloc(gap->grp_ia,(gap->grp_iasize + 1) * (sizeof newsitm));
                }
              gap->grp_ia[++gap->grp_count] = newsitm;
              }
            }
          }
        }
      if (gap->grp_unread > gap->grp_count) gap->grp_unread = gap->grp_count;
      }
    }
  else {
    key[0] = 0;
    if (count && gap->grp_reg_text && loc_disp_us) {
      char *cp;
      int funr;

      if ((cp = strchr(gap->grp_reg_text,'<')) && (sscanf(cp+1,"%d",&funr) == 1)) {
        funr++;
        funr -= loc_disp_stk;
        if (funr <0) funr = 0;
        key[0] = funr;
        }
      }
    key[1] = gap->grp_num;
    itmrab.rab$l_kbf = key;
    itmrab.rab$b_ksz = 8;
    itmrab.rab$b_krf = 0;
    itmrab.rab$l_rop = RAB$M_KGE | RAB$M_RRL | RAB$M_NLK;
    itmrab.rab$b_rac = RAB$C_KEY;

    while (sys$get(&itmrab) & 1) {
      if (newsitm.itm_grp != key[1]) {
	if (gap->grp_count) {
	  if (gap->grp_ia[gap->grp_count-1].itm_num > gap->grp_topnum)
	    gap->grp_topnum = gap->grp_ia[gap->grp_count-1].itm_num;
	  }
	break;
	}
      itmrab.rab$b_rac = RAB$C_SEQ;
      itmrab.rab$l_rop = RAB$M_RRL | RAB$M_NLK;
      if (gap->grp_iasize == gap->grp_count) {
        gap->grp_iasize += 10;
        gap->grp_ia = realloc(gap->grp_ia,(gap->grp_iasize + 1) * (sizeof newsitm));
        }
      newsitm.itm_flags &= ~NEWS_M_NOACCESS;
      if (!(gap->grp_flags & NEWS_M_NNTPSRV))
        newsitm.itm_flags |= NEWS_M_LINESVALID;
      newsitm.itm_flags |= NEWS_M_UNREAD;
      newsitm.itm_title[SUBJLEN-1] = '\0';	/* temporary */
      newsitm.itm_id[IDLEN-1] = '\0';		/* temporary */
      gap->grp_ia[++gap->grp_count] = newsitm;
      }
    }
  gap->grp_unread = gap->grp_count;
  l_ia = gap->grp_ia;
  if ((itm_list = gap->grp_reg_text) != 0) {
    char *cp, *cp1;
    while (itm_list) {
      if ((cp = strchr(itm_list,',')) != 0) {
        while (itm_list) {
	  cp = chop_str_plus(itm_list,',');
          if ((cp1 = strchr(itm_list,'-')) != 0) {
            sscanf(itm_list,"%d-%d",&f_itm,&l_itm);
            for (i = f_itm; i <= l_itm; ++i)
              if (find_itm_by_num(g,i,&j)) l_ia[j].itm_flags &= ~NEWS_M_UNREAD;
            }
	  else {
            sscanf(itm_list,"%d",&f_itm);
            if (find_itm_by_num(g,f_itm,&i)) l_ia[i].itm_flags &= ~NEWS_M_UNREAD;
	    }
          itm_list = cp;
          }
        }
      else if ((*itm_list == '<') || (*itm_list == '|')) {
        if (*itm_list == '|') {
          sscanf(++itm_list,"%d",&f_itm);
          itm_list = strchr(itm_list,'=');
          }
        else f_itm = max(ga[g]->grp_firstnum,1);
        sscanf(++itm_list,"%d",&l_itm);
        itm_list = strchr(itm_list,' ');
        if (itm_list) ++itm_list;
        for (i = f_itm; i <= l_itm; ++i)
          if (find_itm_by_num(g,i,&j)) l_ia[j].itm_flags &= ~NEWS_M_UNREAD;
        }
      else {
        sscanf(itm_list,"%d",&f_itm);
        itm_list = strchr(itm_list,' ');
        if (itm_list) ++itm_list;
        if (find_itm_by_num(g,f_itm,&i)) l_ia[i].itm_flags &= ~NEWS_M_UNREAD;
        }
      }
    gap->grp_unread = 0;
    if (loc_disp_ui) {
      for (i = 1; i <= gap->grp_count; ++i) {
        if (!(l_ia[i].itm_flags & NEWS_M_UNREAD)) {
          int j = i + 1;

          while (j <= gap->grp_count) {
            l_ia[j-1] = l_ia[j];
            ++j;
            }
          --(gap->grp_count);
          }
        }
      }
    for (i = 1; i <= gap->grp_count; ++i)
      if (l_ia[i].itm_flags & NEWS_M_UNREAD) ++(gap->grp_unread);
    free(gap->grp_reg_text);
    gap->grp_reg_text = 0;
    }
  if ((gap->grp_flags & NEWS_M_ORDERBYSUBJECT) && gap->grp_count) {
    int i, j, k;
    ITM_PTR tmp, ttmp;
    char tmp_title[128], title[256], *cp;

    tmp = malloc((gap->grp_iasize + 1) * (sizeof newsitm));
    j = 0;
    for (i = 1; i <= gap->grp_count; ++i) {
      if (!(gap->grp_ia[i].itm_num)) continue;
      tmp[++j] = gap->grp_ia[i];
      if (i == gap->grp_count) break;
      gap->grp_ia[i].itm_num = 0;
      strcpy(title,tmp[j].itm_title);
      lower_case(title);
      cp = title;
      while (!strncmp(cp,"re:",3)) {
        cp += 3;
    	while ((*cp) && (isspace(*cp))) cp++;
    	}
      if (!*cp) continue;
      strip(cp,strlen(cp));
      if (strlen(cp) > 20) cp[20] = '\0';
      for (k = i + 1; k <= gap->grp_count; ++k) {
        if (!(gap->grp_ia[k].itm_num)) continue;
        strcpy(tmp_title,gap->grp_ia[k].itm_title);
        lower_case(tmp_title);
        if (substrcmp(tmp_title,cp)) {
	  tmp[++j] = gap->grp_ia[k];
          gap->grp_ia[k].itm_num = 0;
          }
        }
      }
    ttmp = gap->grp_ia;
    gap->grp_ia = tmp;
    free(ttmp);
    }
  screen_update_gread(g);
  return 0;
}

#if NNTP_CLIENT
static int nntp_get_linecount(cp1)
  char *cp1;
{
  int indx, line_count, i;

  chop_str(cp1,'\n');
  if (sscanf(cp1,"%d %d",&indx,&line_count) != 2) return 0;

  if (find_itm_by_num(sgi,indx,&i)) {
    sgap->grp_ia[i].itm_lines = line_count;
    sgap->grp_ia[i].itm_flags |= NEWS_M_LINESVALID;
    if (sgap->grp_iavdsize) splashline(&(sgap->grp_ia[i]),&(sgap->grp_iavd),&i);
    }
  return 1;
}
#endif

int ds_lines(g)
  int g;
{
  char *bufptr;
  int count, first, last;

  if (   !g
      || !smg_active
      || !nntp_client
      || !ga[g]->grp_count
      || !ga[g]->grp_ia
      || ga[g]->grp_flags & NEWS_M_MAILGROUP)
    return(0);

#if NNTP_CLIENT
  *nntp_currgroup = '\0';
  sprintf(err_oline,"GROUP %s",ga[g]->grp_name);
  if (nntp_one_call(nntp_node,nntp_proto,err_oline,&bufptr) == 211) {
    count = first = last = 0;
    sscanf(bufptr,"211 %d %d %d",&count,&first,&last);
    sgap = ga[g];
    sgi = g;
    if (count && sgap->grp_count) {
      sprintf(err_oline,"XHDR Lines %d-%d",
	sgap->grp_ia[1].itm_num,
	sgap->grp_ia[sgap->grp_count].itm_num);
      if (!(NNTP_USE_XHDR) || !(nntp_get_info(nntp_node,nntp_proto,err_oline,221,0,nntp_get_linecount))) {
        err_line("NNTP XHDR command not supported on the server");
        }
      }
    }
#endif
  return(0);
}

int do_show_linecounts()
{
  if (   !curr_g
      || news_context <= 1
      || !smg_active
      || !nntp_client
      || !ga[curr_g]->grp_count
      || ga[curr_g]->grp_flags & NEWS_M_MAILGROUP)
    return(0);
  ds_lines(curr_g);
  return(0);
}

#if NNTP_CLIENT
static int nntp_get_from(cp1)
  char *cp1;
{
  int indx, i;

  chop_str(cp1,'\n');
  if (sscanf(cp1,"%d",&indx) != 1) return 0;
  while (isdigit(*cp1)) cp1++;
  while (isspace(*cp1)) cp1++;

  if (find_itm_by_num(sgi,indx,&i)) {
    util_fromcpy(sgap->grp_ia[i].itm_from,cp1);
    if (sgap->grp_iavdsize) splashline(&(sgap->grp_ia[i]),&(sgap->grp_iavd),&i);
    }
  return 1;
}
#endif

int ds_from(g)
  int g;
{
  char *bufptr;
  int count, first, last;

  if (   !g
      || !smg_active
      || !nntp_client
      || !ga[g]->grp_count
      || !ga[g]->grp_ia
      || ga[g]->grp_flags & NEWS_M_MAILGROUP)
    return(0);

#if NNTP_CLIENT
  *nntp_currgroup = '\0';
  sprintf(err_oline,"GROUP %s",ga[g]->grp_name);
  if (nntp_one_call(nntp_node,nntp_proto,err_oline,&bufptr) == 211) {
    count = first = last = 0;
    sscanf(bufptr,"211 %d %d %d",&count,&first,&last);
    sgap = ga[g];
    sgi = g;
    if (count && sgap->grp_count) {
      sprintf(err_oline,"XHDR From %d-%d",
	sgap->grp_ia[1].itm_num,
	sgap->grp_ia[sgap->grp_count].itm_num);
      if (!(NNTP_USE_XHDR) || !(nntp_get_info(nntp_node,nntp_proto,err_oline,221,0,nntp_get_from))) {
        err_line("NNTP XHDR command not supported on the server");
        }
      }
    }
#endif
  return(0);
}

static int nntp_get_postmark(cp1)
  char *cp1;
{
  int indx, postmark, i;

  chop_str(cp1,'\n');
  if (sscanf(cp1,"%d %d",&indx,&postmark) != 2) return 0;

  if (find_itm_by_num(sgi,indx,&i)) {
    sgap->grp_ia[i].itm_recvdate = postmark;
    if (sgap->grp_iavdsize) splashline(&(sgap->grp_ia[i]),&(sgap->grp_iavd),&i);
    }
  return 1;
}

static int nntp_get_date(cp1)
  char *cp1;
{
  char *cp2;
  int indx = 0, postmark, i;

  chop_str(cp1,'\n');
  cp2 = cp1;
  while (isspace(*cp2)) ++cp2;
  while (isdigit(*cp2)) indx = (10 * indx) + *cp2++ - '0';
  while (isspace(*cp2)) ++cp2;
  postmark = parse_usenet_date(cp2);

  if (!indx || !postmark) return 0;

  if (find_itm_by_num(sgi,indx,&i)) {
    sgap->grp_ia[i].itm_recvdate = postmark;
    if (sgap->grp_iavdsize) splashline(&(sgap->grp_ia[i]),&(sgap->grp_iavd),&i);
    }
  return 1;
}

int ds_postmark(g)
  int g;
{
  char *bufptr;
  int count, first, last;

  if (   !g
      || !smg_active
      || !nntp_client
      || !ga[g]->grp_count
      || !ga[g]->grp_ia
      || ga[g]->grp_flags & NEWS_M_MAILGROUP)
    return(0);

  *nntp_currgroup = '\0';
  sprintf(err_oline,"GROUP %s",ga[g]->grp_name);
  if (nntp_one_call(nntp_node,nntp_proto,err_oline,&bufptr) == 211) {
    count = first = last = 0;
    sscanf(bufptr,"211 %d %d %d",&count,&first,&last);
    sgap = ga[g];
    sgi = g;
    if (count && sgap->grp_count) {
      if (nntp_anu_news_server) {
        sprintf(err_oline,"XHDR Postmark-Local %d-%d",
	  sgap->grp_ia[1].itm_num,
	  sgap->grp_ia[sgap->grp_count].itm_num);
        if (!(NNTP_USE_XHDR) || !nntp_get_info(nntp_node,nntp_proto,err_oline,221,0,nntp_get_postmark)) {
          err_line("NNTP XHDR command not supported on the server");
          }
        }
      else {
        sprintf(err_oline,"XHDR Date %d-%d",
	  sgap->grp_ia[1].itm_num,
	  sgap->grp_ia[sgap->grp_count].itm_num);
        if (!(NNTP_USE_XHDR) || !nntp_get_info(nntp_node,nntp_proto,err_oline,221,0,nntp_get_date)) {
          err_line("NNTP XHDR command not supported on the server");
          }
        }
      }
    }
  return(0);
}

int do_show_postmarks()
{
  if (   !curr_g
      || news_context <= 1
      || !smg_active
      || !nntp_client
      || !ga[curr_g]->grp_count
      || ga[curr_g]->grp_flags & NEWS_M_MAILGROUP)
    return(0);
  ds_postmark(curr_g);
  return(0);
}

#if NNTP_CLIENT
/*
 *  map_ids - fill in message-id fields in ia and/or id-itmnum map list
 *            This will only do real work if we're an nntp client and
 *            using XHDR commands, otherwise the id fields were filled in
 *            by map_items(), and we'll simply return here.
 */

static int map_g;

static int additmid(xhdr_line)
  char *xhdr_line;
{
  int itmnum,iidx;
  char *id;

  chop_str(xhdr_line,'\n');
  if (sscanf(xhdr_line,"%d",&itmnum) != 1) return(0);
  if (!(id = strchr(xhdr_line,'>'))) return(0);
  *(++id) = '\0';
  if (!(id = strchr(xhdr_line,'<'))) return(0);
  if (find_itm_by_num(map_g,itmnum,&iidx))
    util_idcpy(ga[map_g]->grp_ia[iidx].itm_id,id);
  else {
    IDMAP_PTR tmp;

    if (!(tmp = malloc(sizeof(IDMAP)))) {
      sprintf(err_oline,"malloc() failure expanding Message-ID list");
      err_line(err_oline);
      return(0);
      }
    tmp->itm_num = itmnum;
    tmp->f_link = 0;
    util_idcpy(tmp->id,id);
    if (!ga[map_g]->grp_idhead) {
      tmp->b_link = 0;
      ga[map_g]->grp_idhead = tmp;
      ga[map_g]->grp_idtail = tmp;
      }
    else {
      (ga[map_g]->grp_idtail)->f_link = tmp;
      tmp->b_link = ga[map_g]->grp_idtail;
      ga[map_g]->grp_idtail = tmp;
      }
    }
  return(1);
}

int map_ids(g)
  unsigned int g;
{
  char *nntp_line;

  if (!ga[g]->grp_ia) map_items(g);
  if (!ga[g]->grp_ia) return(0);
  if (ga[g]->grp_flags & NEWS_M_ITMIDSSET) return(1);

#if NNTP_USE_XHDR
  /* we should only get this far if we're an nntp client and we used the
     XHDR command in map_items() */
  if (strcmp(ga[g]->grp_name,nntp_currgroup)) {
    sprintf(err_oline,"GROUP %s",ga[g]->grp_name);
    if (nntp_one_call(nntp_node,nntp_proto,err_oline,&nntp_line) != 211)
      return(0);
    strcpy(nntp_currgroup,ga[g]->grp_name);
    }
  map_g = g;
  sprintf(err_oline,"XHDR message-id %d-%d",ga[g]->grp_firstnum,
          ga[g]->grp_topnum);
  if (!nntp_get_info(nntp_node,nntp_proto,err_oline,221,0,additmid)) {
    sprintf(err_oline,"NNTP XHDR error reading message ids");
    err_line(err_oline);
    return(0);
    }
  ga[g]->grp_flags |= NEWS_M_ITMIDSSET;
#endif /* NNTP_USE_XHDR */

  return(1);
}
#endif /* NNTP_CLIENT */

/*
 *  mem_reset
 *
 *  reset the memory status to match the current file status
 */

void
mem_reset(cd)
  int cd;
{
  int g,
      num_grps,
      mfo = mailfile_open,
      ssa;


  if (mfo) do_close_mail();
  set_level(1);
  ssa = smg_active;
  write_reg_file();

  for (g = 1; g <= ga_size; ++g) {
    if (ga[g]->grp_ia) free(ga[g]->grp_ia);
    if (ga[g]->grp_iavdsize) smg$delete_virtual_display(&ga[g]->grp_iavd);
    if (ga[g]->grp_idtail) {
      while ((ga[g]->grp_idtail = (ga[g]->grp_idtail)->b_link))
        free((ga[g]->grp_idtail)->f_link);
      free(ga[g]->grp_idhead);
      }
    free(ga[g]);
    }
  free(ga);

  curr_g = 0;
  open_groups(cd);
  if (gv_size) {
    num_grps = ga[0]->grp_count;
    if (num_grps > gv_size)
    c$cks(smg$change_virtual_display(&grp_vd,&num_grps,c$ac(devcol)));
    gv_size = num_grps;
    }
  if (ssa && !screen_displays_setup) {
    smg_active = 0;
    init_screen();
    }
  else if (smg_active) screen_map_dir();
  no_more_news = 0;
  if (mfo) do_open_mail();
}

/*
 *  noscreen
 *
 *  Revert to line mode interface
 */

int noscreen()
{
    int g;

    if (smg_active) {
        smg$begin_pasteboard_update(&pid);
        no_broad_trap();
        set_level(1);
        for (g = 1; g <= ga_size; ++g) {
            if (ga[g]->grp_iavdsize) {
                smg$delete_virtual_display(&ga[g]->grp_iavd);
                ga[g]->grp_iavdsize = 0;
                }
            }
        smg$delete_virtual_display(&grp_vd);
        gv_size = 0;
        smg$delete_virtual_display(&grp_header_vd);
        smg$delete_virtual_display(&itm_header_vd);
        smg$delete_virtual_display(&trailer_vd);
        smg$delete_virtual_display(&(itmptr.sdid));
        smg$delete_virtual_display(&(itmptr.shdid));
        smg$delete_virtual_display(&(textptr.sdid));
        smg$delete_virtual_display(&(textptr.shdid));
        while (smg$end_pasteboard_update(&pid) == SMG$_BATSTIPRO);
        reset_mode();
        smg_active = 0;
        screen_displays_setup = 0;
        }
    return(0);
}

int do_noscreen()
{
  return(unwind_display(OUTER_LOOP,noscreen));
}

/*
 *  position_display
 *
 *  reposition the virtual display vd (currently pasted at *cp) to display
 *  line p
 */

void
position_display(vd,cp,p,refresh,maxsize)
    int *vd,
        *cp,
        p,
        refresh,
        maxsize;
{
    if (p <= 7) {
        if (*cp == 4) return;
        *cp = 4;	/* display 1st line */
        }
    else if (p > (maxsize - 7)) {
        if (*cp <= ((devrow - 2) - maxsize)) return;
        *cp = ((devrow - 2) - maxsize);	    /* display last line */
        }
    else {
        if ((*cp >= (11 - p)) && (*cp <= ((devrow - 8) - p))) return;
        if (*cp < (11 - p)) *cp = (11 - p);
        else *cp = ((devrow - 8) - p);
        }
/* OLD CODE-------------------------
    if ((maxsize - p < 7) || (p < 7)) {
        if ((p >= (5 - *cp)) && (p <= (devrow - (2 + *cp)))) return;
        if (p < (5 - *cp)) *cp = (5 - p);
        else *cp = (devrow - (2 + p));
        }
    else {
        if ((p >= (11 - *cp)) && (p <= (devrow - (9 + *cp)))) return;
        if (p < (11 - *cp)) *cp = (11 - p);
        else *cp = (devrow - (9 + p));
        }
----------------------------------------*/
    if (refresh) c$cks(smg$end_pasteboard_update(&pid));
    smg$move_virtual_display(vd,&pid,cp,c$ac(1),0);
    if (refresh) c$cks(smg$begin_pasteboard_update(&pid));
}

/*
 *  screen_map_dir
 *
 *  Redraw the group level screen virtual display
 */

void
screen_map_dir()
{
  int g,
      sg = 0;
  char dir_type[120];

  g_arrow = 0;
  smg$erase_display(&grp_vd);
  if ((!curr_g) && ga_size) curr_g = 1;

  for (g = 1; g <= ga_size; ++g) {
    if (ga[g]->grp_display_indx) {
      ga[g]->grp_display_indx = ++sg;

      if (nntp_client && ga[g]->grp_firstnum < 1)
          sprintf(err_oline,"   %-5d %c %-*.*s           %5d",sg,
              ((ga[g]->grp_flags & NEWS_M_NNTPSRV)? '_' : ' '),SUBJLEN,SUBJLEN,
              ga[g]->grp_name,ga[g]->grp_unread);
      else sprintf(err_oline,"   %-5d %c %-*.*s  %5d    %5d",sg,
              ((ga[g]->grp_flags & NEWS_M_NNTPSRV)? '_' : ' '),SUBJLEN,SUBJLEN,
              ga[g]->grp_name,ga[g]->grp_count,ga[g]->grp_unread);
      smg$put_chars(&grp_vd,c$dscl(&err_oline_d),&sg,c$ac(1),0,0,0,0);
      if (g == curr_g) {
        g_arrow = sg;
        smg$put_chars(&grp_vd,c$dsc("->"),&sg,c$ac(1),0,c$ac(SMG$M_REVERSE+SMG$M_BOLD),0,0);
        }
      if (ga[g]->grp_reg) {
        smg$change_rendition(&grp_vd,&sg,c$ac(12),c$ac(1),c$ac(SUBJLEN),c$ac(SMG$M_BOLD),0);
        if (ga[g]->grp_unread) smg$change_rendition(&grp_vd,&sg,c$ac(4),c$ac(1),c$ac(5),c$ac(SMG$M_BOLD),0);
        }
      }
    }
  grp_display_size = sg;
  if (cur_dir_type == DIR_DATE) sprintf(dir_type,"NEWSGROUPS %s [SINCE=%s, %d Newsgroups]",class_name(),dir_sincestr,sg);
  else if (cur_dir_type == DIR_NEW) sprintf(dir_type,"NEWSGROUPS %s [NEW, %d Newsgroups]",class_name(),sg);
  else if (cur_dir_type == DIR_REGISTER)  sprintf(dir_type,"NEWSGROUPS %s [REGISTERED, %d Newsgroups]",class_name(),sg);
  else if (cur_dir_type == DIR_UNREGISTER)  sprintf(dir_type,"NEWSGROUPS %s [UNREGISTERED, %d Newsgroups]",class_name(),sg);
  else sprintf(dir_type,"NEWSGROUPS %s [ALL, %d Newsgroups]",class_name(),sg);
  smg$put_chars(&grp_header_vd,c$dsc(dir_type),c$ac(1),c$ac(4),c$ac(1),0,0,0);

  if ((!g_arrow) || (grp_display_size <= devrow - 6)) grp_paste = 4;
  else {
    grp_paste = 11 - g_arrow;
    if (grp_paste > 4) grp_paste = 4;
    if (grp_paste < ((devrow - 2) - grp_display_size)) grp_paste = (devrow - 2) - grp_display_size;
    }
  smg$move_virtual_display(&grp_vd,&pid,&grp_paste,c$ac(1),0);
}

void newsgroup_status_line(g)
  unsigned int g;
{
  GRP_PTR gap = ga[g];
  char itemhold[40];

  gwrite_access(g);
  *err_oline = '\0';
  if (gap->grp_reg) {
    sprintf(itemhold,"Reg:%u ",gap->grp_reg);
    strcat(err_oline,itemhold);
    }

  if (gap->grp_flags & NEWS_M_RESTRICT_SET) strcat(err_oline,"Restrict ");

  if (!nntp_client) {
    strcat(err_oline,"Prot:R");
    if (gap->grp_flags & NEWS_M_WRITE_ACCESS) strcat(err_oline,"W");
    if (gap->grp_flags & NEWS_M_MOD_ACCESS) strcat(err_oline,"M");
    else if (gap->grp_flags & NEWS_M_MOD_USER) strcat(err_oline,"m");
    if (!no_priv()) strcat(err_oline,"X");
    }

  if (nntp_client) {
    strcat(err_oline," Srv:");
    strcat(err_oline,nntp_node);
    }
  else if (ga[g]->grp_flags & NEWS_M_NNTPSRV) {
    strcat(err_oline," Srv:");
    if (gap->grp_srvproto) strcat(err_oline,"@");
    strcat(err_oline,gap->grp_srvnode);
    if (!gap->grp_srvproto) strcat(err_oline,"::");
    if (gap->grp_flags & NEWS_M_NNTPCACHE)
    sprintf(&err_oline[strlen(err_oline)],",(cache:%d)",gap->grp_srvcache);
    }
}

void newsgroup_header()
{
  char tmp[IO_SIZE];
  GRP_PTR gap = ga[curr_g];
  ITM_PTR iap;

  newsgroup_status_line(curr_g);
  if (!gap->grp_ia) map_items(curr_g);
  iap = gap->grp_ia;
  if (gap->grp_count)
    sprintf(tmp,"%s: %d Items (#%d - #%d) %s",gap->grp_name,gap->grp_count,
        iap[1].itm_num,iap[gap->grp_count].itm_num,err_oline);
  else sprintf(tmp,"%s: %d Items %s",gap->grp_name,gap->grp_count,err_oline);
  smg$put_chars(&itm_header_vd,c$dsc(tmp),c$ac(1),c$ac(1),c$ac(1),c$ac(SMG$M_BOLD),0,0);
  if (*(gap->grp_topic)) {
    sprintf(tmp,"        -<%s>-",gap->grp_topic);
    smg$put_chars(&itm_header_vd,c$dsc(tmp),c$ac(2),c$ac(1),c$ac(1),c$ac(SMG$M_BOLD),0,0);
    }
  else smg$erase_line(&itm_header_vd,c$ac(2),c$ac(1));
  smg$erase_line(&itm_header_vd,c$ac(3),c$ac(1));
  if (*(gap->grp_notice)) {
    sprintf(tmp,"         [Notice: %s]",gap->grp_notice);
    smg$put_chars(&itm_header_vd,c$dsc(tmp),c$ac(3),c$ac(1),c$ac(1),c$ac(SMG$M_BOLD),0,0);
    }
  else {
    int li, ri, i;

    smg$put_chars(&itm_header_vd,c$dsc("      Title"),c$ac(3),c$ac(4),0,0,0,0);
    if (minfromlen > 0 && !v59_file && devcol >= 80) {
      li = (67 - min(minfromlen,devcol - 30)) + (devcol - 80);
      ri = SUBJLEN + 9;
      i = min(li,ri);
      smg$put_chars(&itm_header_vd,c$dsc(" From"),c$ac(3),c$ac(i),0,0,0,0);
      li = devcol - 6;
      ri = i + minfromlen + 6;
      }
    else {
      li = devcol - 6;
      ri = SUBJLEN + 15;
      }
    i = max(li,ri);
    smg$put_chars(&itm_header_vd,c$dsc("  Date"),c$ac(3),c$ac(i),0,0,0,0);
    i -= 6;
    smg$put_chars(&itm_header_vd,c$dsc(" Lines"),c$ac(3),c$ac(i),0,0,0,0);
    }
}

/*
 *  set_level
 *
 *  Set the screen display level
 */

void set_level(l)
  int l;
{
  int i;
  GRP_PTR gap = ga[curr_g];
  ITM_PTR iap;

  if (news_context == 3) {
    news_context = 2;
    if (smg_active) {
      smg_active = 2;
      if (itmptr.text_displayed) {
        smg$unpaste_virtual_display(&(itmptr.sdid),&pid);
        smg$unpaste_virtual_display(&(itmptr.shdid),&pid);
        itmptr.text_displayed = 0;
        }
      if (textptr.text_displayed) {
        smg$unpaste_virtual_display(&(textptr.sdid),&pid);
        smg$unpaste_virtual_display(&(textptr.shdid),&pid);
        textptr.text_displayed = 0;
        }
      }
    }
  if (news_context == l) return;
  if (l == 1) {
    news_context = 1;
    curr_i = -1;
    if (smg_active) {
      smg_active = 1;
      if ((!curr_g) || (!gap->grp_iavdsize)) return;
      smg$unpaste_virtual_display(&itm_header_vd,&pid);
      smg$unpaste_virtual_display(&(gap->grp_iavd),&pid);
      }
    return;
    }

  if (!curr_g) return;
  if (!gap->grp_count) {
    if (smg_active && (!gap->grp_iavdsize)) {
      smg$create_virtual_display(&devrow,c$ac(devcol),&(gap->grp_iavd),0,0,0);
      smg$put_chars(&(gap->grp_iavd),c$dsc("NEWSGROUP is empty"),c$ac(1),c$ac(20),0,c$ac(SMG$M_REVERSE),0,0);
      gap->grp_iavdsize = devrow;
      gap->grp_iapaste = 4;
      }
    gap->grp_c_itm = 0;
    }
  else {
    if (!gap->grp_ia) map_items(curr_g);
    if (!gap->grp_count) {
      if (smg_active && (!gap->grp_iavdsize)) {
        smg$create_virtual_display(&devrow,c$ac(devcol),&(gap->grp_iavd),0,0,0);
        smg$put_chars(&(gap->grp_iavd),c$dsc("NEWSGROUP is empty"),c$ac(1),c$ac(20),0,c$ac(SMG$M_REVERSE),0,0);
        gap->grp_iavdsize = devrow;
        gap->grp_iapaste = 4;
        }
      gap->grp_c_itm = 0;
      }
    else {
      iap = gap->grp_ia;
      if (smg_active && (!gap->grp_iavdsize)) {
        smg$create_virtual_display(c$rfi(gap->grp_iavdsize = max(devrow,gap->grp_count)),c$ac(devcol),&(gap->grp_iavd),0,0,0);
        for (i = 1; i <= gap->grp_count; ++i) {
          splashline(&(iap[i]),&(gap->grp_iavd),&i);
          if (iap[i].itm_flags & NEWS_M_UNREAD) {
            if (!gap->grp_c_itm) {
              gap->grp_c_itm = i;
              smg$put_chars(&gap->grp_iavd,c$dsc("->"),&i,c$ac(1),0,c$ac(SMG$M_REVERSE+SMG$M_BOLD),0,0);
              }
            }
          }
        }
      if (!gap->grp_c_itm) {
        for (i = 1; i < gap->grp_count; ++i)
          if (iap[i].itm_flags & NEWS_M_UNREAD) break;
        gap->grp_c_itm = i;
        if (gap->grp_iavdsize) smg$put_chars(&gap->grp_iavd,c$dsc("->"),&i,c$ac(1),0,c$ac(SMG$M_REVERSE+SMG$M_BOLD),0,0);
        }
      }
    }
  curr_i = gap->grp_c_itm;
  if (smg_active) {

    gap->grp_iapaste = 4;
    if (curr_i) {
      position_display(&gap->grp_iavd,&gap->grp_iapaste,gap->grp_count,0,gap->grp_count);
      position_display(&gap->grp_iavd,&gap->grp_iapaste,curr_i,0,gap->grp_count);
      smg$put_chars(&gap->grp_iavd,c$dsc("->"),&curr_i,c$ac(1),0,c$ac(SMG$M_REVERSE+SMG$M_BOLD),0,0);
      }
    newsgroup_header();
    c$cks(smg$paste_virtual_display(&(gap->grp_iavd),&pid,&(gap->grp_iapaste),c$ac(1)));
    c$cks(smg$paste_virtual_display(&itm_header_vd,&pid,c$ac(1),c$ac(1)));
    c$cks(smg$repaste_virtual_display(&trailer_vd,&pid,c$rfi(devrow - 2),c$ac(1)));
    smg_active = 2;
    }
  news_context = 2;
}

/*
 *  util_cvrt
 *
 *  Convert a string into standard newsgroup format
 */

void util_cvrt(result, input)
    char *result,
         *input;
{
    char *p = result,
         *in = input;
    int i;
    while ((*in) && (*in == ' ')) in++;
    strncpy(result,in,SUBJLEN);
    result[SUBJLEN - 1] = '\0';
    strip(result,strlen(result));
    while (*p) {
        if (isgraph(*p))
#if vax11c
	  *p = tolower(*p);
#else
	  {
          if (*p >= 'A' && *p <= 'Z')
             *p += 'a' - 'A' ;
          }
#endif
        else
          *p = '_';
        p++;
        }
    i = strlen(result);
    while (i < SUBJLEN) result[i++] = '\0';
}

/*
 *  util_dir
 *
 *  Convert a newsgroup name to a VMS directory string
 */

char dir_result[SUBJLEN + SUBJLEN];

char *
util_dir(input)
    char *input;
{
    char *p = dir_result,
         *in = input;

    while (*in) {
        if (isalnum(*in) || (*in == '-') || (*in == '.')) *p++ = *in++;
        else {
            *p++ = '_';
            if (*in == '_') *p++ = '_';
            else if (*in < '0') *p++ = (*in - '!') + 'A';
            else if (*in < 'A') *p++ = (*in - ':') + '0';
            else if (*in < 'a') *p++ = (*in - '[') + 'P';
            else *p++ = (*in - '{') + 'V';
            in++;
            }
        }
    *p = '\0';
    return(dir_result);
}

/*
 *  util_undir
 *
 *  Convert a directory name to a newsgroup string
 */

char undir_result[SUBJLEN];

char *
util_undir(input)
    char *input;
{
    char *p = undir_result,
         *in = input;

    while (*in) {
        if (*in == '_') {
            in++;
            if (*in < 'A') *p++ = (*in - '0') + ':';
            else if (*in < 'P') *p++ = (*in - 'A') + '!';
            else if (*in < 'V') *p++ = (*in - 'P') + '[';
            else *p++ = (*in - 'V') + '{';
            }
        else *p++ = tolower(*in);
        in++;
        }
    *p = '\0';
    return(undir_result);
}

/*
 *  strip
 *
 *  remove trailing blanks
 */

void
strip(p,n)
    char *p;
    int n;
{
    do {
        n--;
        } while ((n > 0) && (p[n] == ' '));
    p[n+1] = '\0';
}

/*
 *  do_refresh
 *
 *  rewrite the SMG screen
 */

int do_refresh()
{
  if (smg_active) {
    smg$end_pasteboard_update(&pid);
    smg$repaint_screen(&pid);
    smg$begin_pasteboard_update(&pid);
    smg$set_keypad_mode(&kid,c$ac(SMG$M_KEYPAD_APPLICATION));
    }
  return(0);
}

/*
 *  do_error
 *
 *  CLI problem
 */

int do_error()
{
  err_line("Error: News - unrecognised command");
  return(0);
}

/*
 *  do_exit
 *
 *  bye for now
 */

int do_exit()
{
    return(1);
}

/*
 *  do_quit
 *
 *  Close rms files, DONT write the register file, shutdown the screen, and
 *  send any printer output to a nominated queue.
 */

int do_quit()
{
    err_line("QUIT will NOT save any REGISTER, READ, MARK, SKIP commands from this session");
    status = get_input(&usr_inp_dsc,c$dsc("Quit? [y]"),&usr_inp_l);
    if (!(status & 1)) return(1);
    if (!(!usr_inp_l || (*usr_inp == 'Y') || (*usr_inp == 'y'))) return(0);
    sysprv();
    sys$close(&itmfab);
    sys$close(&grpfab);
    nosysprv();
    close_hist_file();
    close_nntp_file();
    close_mail_file();
    noscreen();
    print_exit();
    log_to_usage_file(1,0);
    exit(1);
    /*NOTREACHED*/
    return 1;
}

/*
 *  screen_grp_display
 *
 *  NEWSGROUPS command - set the screen to the newsgroup directory
 */

int screen_grp_display()
{
    set_level(1);
    return(0);
}

int do_groups_display()
{
  return(unwind_display(OUTER_LOOP,screen_grp_display));
}

/*
 *  screen_cur_down
 *
 *  Move the cursor down
 */

int screen_cur_down()
{
  char line[132];
  unsigned short line_len;
  int lines;
  $DESCRIPTOR(line_dsc,line);

  if (cli$get_value(c$dsc("LINES"),&line_dsc,&line_len) & 1) {
    line[line_len] = '\0';
    if (sscanf(line,"%d",&lines) != 1) lines = 1;
    }
  else lines = 1;
  if (!lines) return(0);
  if (lines < 0) {
    if (news_context == 1) cur_up_grp(lines,1);
    else {
      set_level(2);
      cur_up_itm(curr_g,lines,1);
      }
    return(0);
    }
  if (news_context == 1) cur_down_grp(lines,1);
  else {
    set_level(2);
    cur_down_itm(curr_g,lines,1);
    }
  return(0);
}

/*
 *  screen_cur_up
 *
 *  move the cursor up by n lines on the screen
 */

int screen_cur_up()
{
    char line[132];
    unsigned short line_len;
    int lines;
    $DESCRIPTOR(line_dsc,line);

    if (cli$get_value(c$dsc("LINES"),&line_dsc,&line_len) & 1) {
        line[line_len] = '\0';
        if (sscanf(line,"%d",&lines) != 1) lines = 1;
        }
    else lines = 1;
    if (!lines) return(0);
    if (lines < 0) {
        if (news_context == 1) cur_down_grp(lines,1);
        else {
            set_level(2);
            cur_down_itm(curr_g,lines,1);
            }
        return(0);
        }
    if (news_context == 1) cur_up_grp(lines,1);
    else {
        set_level(2);
        cur_up_itm(curr_g,lines,1);
        }
    return(0);
}

int do_top()
{
  if (news_context == 1) cur_up_grp(1000000,1);
  else {
    set_level(2);
    cur_up_itm(curr_g,1000000,1);
    }
  return(0);
}

int do_bottom()
{
  if (news_context == 1) cur_down_grp(1000000,1);
  else {
    set_level(2);
    cur_down_itm(curr_g,1000000,1);
    }
  return(0);
}

int check_id(id)
  char *id;
{
  char mref[IDLEN + 4];
  int *gp;

  util_idcpy(mref,id);
  gp = (int *)&mref[IDLEN];
  *gp = 0;

  itmrab.rab$l_kbf = mref;
  itmrab.rab$b_ksz = IDLEN + 4;
  itmrab.rab$b_krf = 1;
  itmrab.rab$l_rop = RAB$M_KGE | RAB$M_RRL | RAB$M_NLK;
  itmrab.rab$b_rac = RAB$C_KEY;

  if (   (sys$get(&itmrab) & 1)
      && !strcmp(mref,newsitm.itm_id)) return(1);
  if (hist_check(mref)) return(1);
  return(0);
}

int
do_noop()
{
  return(0);
}

static char *markoff = 0;

static char *regtext_array(first,last,text)
  int first, last;
  char *text;
{
  int f_itm, l_itm,i;
  char *cp = text;

  if (markoff) free(markoff);
  markoff = 0;
  if (last < first) return(0);
  markoff = malloc((last - first) + 1);
  for (i = first; i <= last; ++i) markoff[i-first] = 1;
  if (!text || !*text) return(markoff);
  while (*cp) {
    if ((*cp == '<') || (*cp == '|')) {
      if (*cp == '|') {
        sscanf(++cp,"%d",&f_itm);
        cp = strchr(cp,'=');
        }
      else f_itm = 1;
      sscanf(++cp,"%d",&l_itm);
      cp = strchr(cp,' ');
      if (cp) ++cp;
      if (f_itm < first) f_itm = first;
      if (l_itm > last) l_itm = last;
      for (i = f_itm; i <= l_itm; ++i) markoff[i-first] = 0;
      }
    else {
      sscanf(cp,"%d",&f_itm);
      cp = strchr(cp,' ');
      if (cp) ++cp;
      if ((f_itm >= first) && (f_itm <= last)) markoff[f_itm - first] = 0;
      }
    }
  return(markoff);
}

static char *ret_range = 0;

static char *regtext_range(first,last,text)
  int first, last;
  char *text;
{
  int i, f_itm = 0, l_itm = 0,rr_len = 0, rt_len = 0;
  char *mask, tmptxt[80];

  if (ret_range) free(ret_range);
  ret_range = 0;
  if (!(mask = regtext_array(first,last,text))) return(0);

  for (i = first; i <= last; ++i) {
    if (!mask[i-first]) {
      if (f_itm) {
        if (l_itm == f_itm) sprintf(tmptxt,"%d",f_itm);
        else sprintf(tmptxt,"%d-%d",f_itm,l_itm);
        if ((rt_len + strlen(tmptxt) + 10) > rr_len) {
          if (!rr_len) ret_range = malloc(rr_len = 256);
          else ret_range = realloc(ret_range,rr_len += 256);
          }
        if (!*ret_range) strcpy(ret_range,tmptxt);
        else {
          strcat(ret_range," ");
          strcat(ret_range,tmptxt);
          }
        rt_len = strlen(ret_range);
        f_itm = l_itm = 0;
        }
      }
    else {
      if (!f_itm) f_itm = l_itm = i;
      else l_itm = 1;
      }
    }
  if (f_itm) {
    if (l_itm == f_itm) sprintf(tmptxt,"%d",f_itm);
    else sprintf(tmptxt,"%d-%d",f_itm,l_itm);
    if ((rt_len + strlen(tmptxt) + 10) > rr_len) {
      if (!rr_len) ret_range = malloc(rr_len = 256);
      else ret_range = realloc(ret_range,rr_len += 256);
      }
    if (!*ret_range) strcpy(ret_range,tmptxt);
    else {
      strcat(ret_range," ");
      strcat(ret_range,tmptxt);
      }
    }
  return(ret_range);
}

int scdown()
{
  int lines = devrow - 6;

  if (news_context == 1) cur_down_grp(lines,1);
  else {
    set_level(2);
    cur_down_itm(curr_g,lines,1);
    }
  return(0);
}

int scup()
{
  int lines = devrow - 6;

  if (news_context == 1) cur_up_grp(lines,1);
  else {
    set_level(2);
    cur_up_itm(curr_g,lines,1);
    }
  return(0);
}

int
do_abort()
{
    sysprv();
    sys$close(&itmfab);
    sys$close(&grpfab);
    nosysprv();
    close_hist_file();
    close_nntp_file();
    close_mail_file();
    noscreen();
    log_to_usage_file(1,0);
    exit(1);
    /*NOTREACHED*/
    return 0;
}

static char *dat()
{
  static char datum[32];
  time_t cur_time;
  char *p;
  time(&cur_time);
  p = ctime(&cur_time);
  strncpy(datum, p, 19);
  return(datum);
}

void
log_to_usage_file(flag,group)
  int flag;				/*  0 - initialize data areas */
					/*  1 - dump data to logging file */
					/* -1 - read article from group */
					/* -2 - post article to group */
  int group;				/* group number for read/post */
{                                       /* logging declarations */
  FILE *fpl;
  int g, i, retry_cnt;
  long dif_time;
  time_t end_time;
  static time_t start_time;
  static int max_index, read_count, post_count;
  static int *group_nums, *read_groups, *post_groups;
  char *trans;
  unsigned short len = 0;
  if (flag == 0) {
    time(&start_time);
    max_index = 0;
    read_count = 0;
    post_count = 0;
    group_nums = malloc(100 * sizeof(int));
    read_groups = malloc(100 * sizeof(int));
    post_groups = malloc(100 * sizeof(int));
    return;
    }
  if (group) group = ga[group]->grp_num;
  if (flag < 0) {
    for (g = 0; g < max_index; g++)
      if (group_nums[g] == group) break;
    if (g == max_index) {
      if (max_index == 100) g = -1;
      else {
        max_index++;
        group_nums[g] = group;
        read_groups[g] = 0;
        post_groups[g] = 0;
        }
      }

    if (flag == -1) {
      read_count++;
      if (g >= 0) read_groups[g]++;
      return;
      }
    else if (flag == -2) {
      post_count++;
      if (g >= 0) post_groups[g]++;
      return;
      }
    }
  /*  if (!no_priv()) return;  */
                   /*    Translate logical name. */
  if (!(trans = news_getenv("NEWS_USAGE_LOG",1))) return;
  sysprv();
  retry_cnt = 3;
  do {
    if (!(fpl = fopen(trans,"a","shr=get"))) sleep(1); else break;
    } while (--retry_cnt >= 0);
  if (fpl) {
    time(&end_time);
    dif_time = (start_time > end_time) ? 0L : end_time - start_time;
    fprintf(fpl,"%s: User %s - %ld minutes - %d articles read - %d posted\n",
      dat(), usr_username, (dif_time + 30) / 60, read_count, post_count);
    for (i = 0; i < max_index; i++) {
      g = ga_locate(group_nums[i]);
      if (!g) continue;
      fprintf(fpl,"\t%s %d %d\n",ga[g]->grp_name,read_groups[i],post_groups[i]);
      }
    fclose(fpl);
    }
  nosysprv();
}
